#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@file: regulatory_network.py
@time: 2023/Jan/08
@description: inference gene regulatory networks
@author: Yao LI
@email: liyao1@genomics.cn
@last modified by: Yao LI
"""

# python core modules
import os

# third party modules
import anndata
import pandas as pd
import numpy as np
import scanpy as sc
import seaborn as sns
import matplotlib.pyplot as plt
from typing import Optional
from pyscenic.rss import regulon_specificity_scores
import matplotlib as mpl
mpl.rcParams['pdf.fonttype'] = 42
mpl.rcParams['ps.fonttype'] = 42
mpl.rcParams['svg.fonttype'] = 'none'
mpl.rcParams["ytick.labelright"] = True
mpl.rcParams["ytick.labelleft"] = False

# modules in self project


################################################
#                                              #
#        Plot Gene Regulatory Networks         #
#                                              #
################################################
COLORS = [
    '#d60000', '#e2afaf', '#018700', '#a17569', '#e6a500', '#004b00',
    '#6b004f', '#573b00', '#005659', '#5e7b87', '#0000dd', '#00acc6',
    '#bcb6ff', '#bf03b8', '#645472', '#790000', '#0774d8', '#729a7c',
    '#8287ff', '#ff7ed1', '#8e7b01', '#9e4b00', '#8eba00', '#a57bb8',
    '#5901a3', '#8c3bff', '#a03a52', '#a1c8c8', '#f2007b', '#ff7752',
    '#bac389', '#15e18c', '#60383b', '#546744', '#380000', '#e252ff',
]
CM = 1 / 2.54


# dotplot method for anndata
def dotplot_anndata(data: anndata.AnnData,
                    gene_names: list,
                    cluster_label: str,
                    save: bool = True,
                    **kwargs):
    """
    create a dotplot for Anndata object.
    a dotplot contains percent (of cells that) expressed (the genes) and average expression (of genes).

    :param data: gene data
    :param gene_names: interested gene names
    :param cluster_label: label of clustering output
    :param save: if save plot into a file
    :param kwargs: features Input vector of features, or named list of feature vectors
    if feature-grouped panels are desired
    :return: plt axe object
    """
    if isinstance(data, anndata.AnnData):
        return sc.pl.dotplot(data, var_names=gene_names, groupby=cluster_label, save=save, **kwargs)


def plot_2d_reg(data: anndata.AnnData,
                auc_mtx,
                reg_name: str,
                fn: str,
                pos_label='spatial',
                **kwargs):
    """
    Plot genes of one regulon on a 2D map
    :param pos_label:
    :param data:
    :param auc_mtx:
    :param reg_name:
    :param fn:
    :return:
    """
    if '(+)' not in reg_name:
        reg_name = reg_name + '(+)'

    cell_coor = data.obsm[pos_label]
    auc_zscore = cal_zscore(auc_mtx)
    # prepare plotting data
    sub_zscore = auc_zscore[reg_name]
    # sort data points by zscore (low to high), because first dot will be covered by latter dots
    zorder = np.argsort(sub_zscore.values)
    # plot cell/bin dot, x y coor
    sc = plt.scatter(cell_coor[:, 0][zorder],
                     cell_coor[:, 1][zorder],
                     c=sub_zscore.iloc[zorder],
                     marker='.',
                     edgecolors='none',
                     cmap='plasma',
                     lw=0,
                     **kwargs)
    plt.axis("equal")
    plt.box(False)
    plt.axis('off')
    plt.colorbar(sc, shrink=0.35)
    plt.savefig(fn, format='pdf')
    plt.close()


def plot_3d_reg(data: anndata.AnnData,
                auc_mtx,
                reg_name: str,
                fn: str,
                view_vertical=222,
                view_horizontal=-80,
                show_bg=False,
                xscale=1,
                yscale=1,
                zscale=1,
                pos_label='spatial',
                **kwargs):
    """
    Plot genes of one regulon on a 3D map
    :param pos_label:
    :param data:
    :param auc_mtx:
    :param reg_name:
    :param fn:
    :param view_vertical: vertical angle to view to the 3D object
    :param view_horizontal: horizontal angle to view the 3D object
    :param show_bg: if show background
    :param xscale:
    :param yscale:
    :param zscale:
    :return:

    Example:
        plot_3d_reg(data, 'spatial', auc_mtx, 'Zfp354c', view_vertical=30, view_horizontal=-30)
    """
    reg_name = f'reg_name(+)' if '(+)' not in reg_name else reg_name

    # prepare plotting data
    cell_coor = data.obsm[pos_label]
    if isinstance(cell_coor, pd.DataFrame):
        cell_coor = cell_coor.to_numpy()
    sub_zscore = auc_mtx[reg_name]

    from mpl_toolkits.mplot3d import Axes3D
    fig = plt.figure()
    ax = Axes3D(fig)
    sc = ax.scatter(cell_coor[:, 0],
                    cell_coor[:, 1],
                    cell_coor[:, 2],
                    c=sub_zscore,
                    marker='.',
                    edgecolors='none',
                    cmap='plasma',
                    lw=0, **kwargs)
    # set view angle
    ax.view_init(view_vertical, view_horizontal)
    # scale axis
    xlen = cell_coor[:, 0].max() - cell_coor[:, 0].min()
    ylen = cell_coor[:, 1].max() - cell_coor[:, 1].min()
    zlen = cell_coor[:, 2].max() - cell_coor[:, 2].min()
    _xscale = xscale
    _yscale = ylen / xlen * yscale
    _zscale = zlen / xlen * zscale
    ax.get_proj = lambda: np.dot(Axes3D.get_proj(ax), np.diag([_xscale, _yscale, _zscale, 1]))

    if not show_bg:
        plt.box(False)
        plt.axis('off')
    plt.colorbar(sc, shrink=0.35)
    plt.savefig(fn)
    plt.close()


def plot_3d_tf(data: anndata.AnnData,
               auc_mtx,
               reg_name: str,
               fn: str,
               view_vertical=222,
               view_horizontal=-80,
               show_bg=False,
               xscale=1,
               yscale=1,
               zscale=1,
               pos_label='spatial',
               **kwargs):
    """
    Plot genes of one regulon on a 3D map
    :param pos_label:
    :param data:
    :param auc_mtx:
    :param reg_name:
    :param fn:
    :param view_vertical: vertical angle to view to the 3D object
    :param view_horizontal: horizontal angle to view the 3D object
    :param show_bg: if show background
    :param xscale:
    :param yscale:
    :param zscale:
    :return:

    Example:
        plot_3d_reg(data, 'spatial', auc_mtx, 'Zfp354c', view_vertical=30, view_horizontal=-30)
    """
    if '(+)' not in reg_name:
        reg_name = reg_name + '(+)'

    # prepare plotting data
    cell_coor = data.obsm[pos_label]
    if isinstance(cell_coor, pd.DataFrame):
        cell_coor = cell_coor.to_numpy()
    auc_zscore = cal_zscore(auc_mtx)
    sub_zscore = auc_zscore[reg_name]

    from mpl_toolkits.mplot3d import Axes3D
    fig = plt.figure()
    ax = Axes3D(fig)
    sc = ax.scatter(cell_coor[:, 0],
                    cell_coor[:, 1],
                    cell_coor[:, 2],
                    c=sub_zscore,
                    marker='.',
                    edgecolors='none',
                    cmap='plasma',
                    lw=0, **kwargs)
    # set view angle
    ax.view_init(view_vertical, view_horizontal)
    # scale axis
    xlen = cell_coor[:, 0].max() - cell_coor[:, 0].min()
    ylen = cell_coor[:, 1].max() - cell_coor[:, 1].min()
    zlen = cell_coor[:, 2].max() - cell_coor[:, 2].min()
    _xscale = xscale
    _yscale = ylen / xlen * yscale
    _zscale = zlen / xlen * zscale
    ax.get_proj = lambda: np.dot(Axes3D.get_proj(ax), np.diag([_xscale, _yscale, _zscale, 1]))

    if not show_bg:
        plt.box(False)
        plt.axis('off')
    plt.colorbar(sc, shrink=0.35)
    plt.savefig(fn)
    plt.close()


def plot_3d_web(data, auc_mtx, reg_name, prefix='', zscale=1, xscale=1, yscale=1):
    """

    :param data:
    :param auc_mtx:
    :param reg_name:
    :param prefix:
    :param zscale:
    :param xscale:
    :param yscale:
    :return:
    """
    coor = data.obsm['spatial_regis']

    import plotly.graph_objects as go
    from plotly.subplots import make_subplots

    sub_zscore = auc_mtx[reg_name]
    fig = make_subplots(rows=1, cols=1,
                        specs=[[{'type': 'scene'}]],
                        subplot_titles=('regulon'),
                        )
    cell = go.Scatter3d(
        x=coor[:, 0],
        y=coor[:, 1],
        z=coor[:, 2],
        mode='markers',
        marker=dict(
            size=2,
            color=sub_zscore,
            opacity=1,
            colorbar=dict(thickness=20)
        ),
        legendgroup='cell',
        showlegend=True,
    )
    fig.add_trace(cell, row=1, col=1)
    axis = dict(
        showbackground=False,
        showline=True,
        zeroline=False,
        showgrid=True,
        showticklabels=False,
        title='',
    )
    fig.update_layout(
        margin=dict(l=10, r=10, b=0, t=0),
        showlegend=True,
        scene=dict(
            aspectmode='data',
            xaxis=axis,
            yaxis=axis,
            zaxis=axis,
        ),
        paper_bgcolor='rgba(0, 0, 0, 1)',
        plot_bgcolor='rgba(0, 0, 0, 1)',
    )
    # manually force the z-axis to appear twice as big as the other two
    fig.update_layout(scene_aspectmode='manual',
                      scene_aspectratio=dict(x=xscale, y=yscale, z=zscale))
    fig.write_html(f'{prefix}_regulon.html')


def auc_heatmap(data: anndata.AnnData,
                auc_mtx: pd.DataFrame,
                cluster_label: str,
                rss_fn=None,
                topn=5,
                save=True,
                subset=True,
                subset_size=5000,
                fn='clusters_heatmap_top5.pdf',
                legend_fn="rss_celltype_legend.pdf",
                cluster_list=None,
                row_cluster=False,
                col_cluster=True,
                cmap="YlGnBu",
                vmin=-3, vmax=3,
                yticklabels=True, xticklabels=True,
                **kwargs):
    """
    Plot heatmap for Regulon specificity scores (RSS) value
    :param data:
    :param auc_mtx:
    :param cluster_label:
    :param rss_fn:
    :param topn:
    :param save:
    :param subset:
    :param subset_size:
    :param fn:
    :param legend_fn:
    :param cluster_list: list of cluster names one prefer to use
    :param row_cluster:
    :param col_cluster:
    :param cmap:
    :param vmin:
    :param vmax:
    :param yticklabels:
    :param xticklabels:
    :param kwargs:
    :return:

    Example:
        # only plot ['CNS', 'amnioserosa', 'carcass'] clusters and their corresponding top regulons
        auc_heatmap(adata, auc_mtx, cluster_label='celltypes', subset=False,
                    rss_fn='regulon_specificity_scores.txt',
                    cluster_list=['CNS', 'amnioserosa', 'carcass'])
    """
    if subset and len(data.obs) > subset_size:
        fraction = subset_size / len(data.obs)
        # do stratified sampling
        draw_obs = data.obs.groupby(cluster_label, group_keys=False).apply(lambda x: x.sample(frac=fraction))
        # load the regulon_list from a file using the load_signatures function
        cell_order = draw_obs[cluster_label].sort_values()
    else:
        # load the regulon_list from a file using the load_signatures function
        cell_order = data.obs[cluster_label].sort_values()
    celltypes = sorted(list(set(data.obs[cluster_label])))

    # Regulon specificity scores (RSS) across predicted cell types
    if rss_fn is None:  # input rss file has highest priority
        if 'rss' in data.uns:
            rss_cellType = data.uns['rss']
        else:
            rss_cellType = regulon_specificity_scores(auc_mtx, data.obs[cluster_label])
    else:
        rss_cellType = pd.read_csv(rss_fn, index_col=0)
    # Select the top 5 regulon_list from each cell type
    topreg = get_top_regulons(data, cluster_label, rss_cellType, topn=topn)

    if cluster_list is None:
        cluster_list = celltypes.copy()
    colorsd = dict((i, c) for i, c in zip(cluster_list, COLORS))
    colormap = [colorsd[x] for x in cell_order]

    # plot legend
    plot_legend(colorsd, fn=legend_fn)

    # plot z-score
    auc_zscore = cal_zscore(auc_mtx)
    try:
        plot_data = auc_zscore[topreg].loc[cell_order.index]
    except KeyError:
        com_topreg = list(set(topreg).intersection(set(auc_zscore.columns)))
        plot_data = auc_zscore[com_topreg].loc[cell_order.index]
    sns.set(font_scale=1.2)
    g = sns.clustermap(plot_data,
                       annot=False,
                       square=False,
                       linecolor='gray',
                       yticklabels=yticklabels, xticklabels=xticklabels,
                       vmin=vmin, vmax=vmax,
                       cmap=cmap,
                       row_colors=colormap,
                       row_cluster=row_cluster,
                       col_cluster=col_cluster,
                       **kwargs)
    g.cax.set_visible(False)  # set colorbar
    g.ax_heatmap.set_yticks([])
    g.ax_heatmap.set_ylabel('')
    g.ax_heatmap.set_xlabel('')
    if save:
        file_format = os.path.splitext(fn)[1].replace('.','')
        plt.savefig(fn, format=file_format)
    return g


def isr_heatmap(data: anndata.AnnData,
                cluster_label: str,
                isr_mtx=None,
                rss_fn=None,
                topn=5,
                save=True,
                subset=True,
                subset_size=5000,
                fn='clusters_heatmap_top5.pdf',
                legend_fn="rss_celltype_legend.pdf",
                cluster_list=None,
                row_cluster=False,
                col_cluster=True,
                cmap="YlGnBu",
                vmin=-3, vmax=3,
                yticklabels=True, xticklabels=True,
                **kwargs):
    """
    Plot heatmap for Regulon specificity scores (RSS) value
    :param data:
    :param cluster_label:
    :param isr_mtx:
    :param rss_fn:
    :param topn:
    :param save:
    :param subset:
    :param subset_size:
    :param fn:
    :param legend_fn:
    :param cluster_list:
    :param row_cluster:
    :param col_cluster:
    :param cmap:
    :param vmin:
    :param vmax:
    :param yticklabels:
    :param xticklabels:
    :param kwargs:
    :return:

    Example:
        # only plot ['CNS', 'amnioserosa', 'carcass'] clusters and their corresponding top regulons
        auc_heatmap(adata, auc_mtx, cluster_label='celltypes', subset=False,
                    rss_fn='regulon_specificity_scores.txt',
                    cluster_list=['CNS', 'amnioserosa', 'carcass'])
    """
    if isr_mtx is None:
        isr_mtx = data.obsm['isr']
    if subset and len(data.obs) > subset_size:
        fraction = subset_size / len(data.obs)
        # do stratified sampling
        draw_obs = data.obs.groupby(cluster_label, group_keys=False).apply(lambda x: x.sample(frac=fraction))
        # load the regulon_list from a file using the load_signatures function
        cell_order = draw_obs[cluster_label].sort_values()
    else:
        # load the regulon_list from a file using the load_signatures function
        cell_order = data.obs[cluster_label].sort_values()
    celltypes = sorted(list(set(data.obs[cluster_label])))

    # Regulon specificity scores (RSS) across predicted cell types
    if rss_fn is None:
        rss_cellType = regulon_specificity_scores(isr_mtx, data.obs[cluster_label])
    else:
        rss_cellType = pd.read_csv(rss_fn, index_col=0)
    # Select the top 5 regulon_list from each cell type
    topreg = get_top_regulons(data, cluster_label, rss_cellType, topn=topn)

    if cluster_list is None:
        cluster_list = celltypes.copy()
    colorsd = dict((i, c) for i, c in zip(cluster_list, COLORS))
    colormap = [colorsd[x] for x in cell_order]

    # plot legend
    plot_legend(colorsd, fn=legend_fn)

    # plot z-score
    isr_zscore = cal_zscore(isr_mtx)
    try:
        plot_data = isr_zscore[topreg].loc[cell_order.index]
    except KeyError:
        com_topreg = list(set(topreg).intersection(set(isr_zscore.columns)))
        plot_data = isr_zscore[com_topreg].loc[cell_order.index]
    sns.set(font_scale=1.2)
    g = sns.clustermap(plot_data,
                       annot=False,
                       square=False,
                       linecolor='gray',
                       yticklabels=yticklabels, xticklabels=xticklabels,
                       vmin=vmin, vmax=vmax,
                       cmap=cmap,
                       row_colors=colormap,
                       row_cluster=row_cluster,
                       col_cluster=col_cluster,
                       **kwargs)
    g.cax.set_visible(False)
    g.ax_heatmap.set_yticks([])
    g.ax_heatmap.set_ylabel('')
    g.ax_heatmap.set_xlabel('')
    if save:
        file_format = os.path.splitext(fn)[1].replace('.','')
        plt.savefig(fn, format=file_format)
    return g


def auc_heatmap_uneven(data: anndata.AnnData,
                       auc_mtx: pd.DataFrame,
                       cluster_label: str,
                       rss_fn: str,
                       topn=5,
                       target_celltype: str = 'ventricular-specific CM',
                       save=True,
                       subset=True,
                       subset_size=5000,
                       fn='clusters_heatmap_top5.png',
                       legend_fn="rss_celltype_legend.png",
                       cluster_list=None):
    """
    Plot heatmap for Regulon specificity scores (RSS) value
    :param data:
    :param auc_mtx:
    :param cluster_label:
    :param rss_fn:
    :param topn:
    :param save:
    :param subset:
    :param subset_size:
    :param fn:
    :param legend_fn:
    :param cluster_list: list of cluster names one prefer to use
    :return:

    Example:
        # only plot ['CNS', 'amnioserosa', 'carcass'] clusters and their corresponding top regulons
        auc_heatmap(adata, auc_mtx, cluster_label='celltypes', subset=False,
                    rss_fn='regulon_specificity_scores.txt',
                    cluster_list=['CNS', 'amnioserosa', 'carcass'])
    """
    if subset and len(data.obs) > subset_size:
        fraction = subset_size / len(data.obs)
        # do stratified sampling
        draw_obs = data.obs.groupby(cluster_label, group_keys=False).apply(lambda x: x.sample(frac=fraction))
        # load the regulon_list from a file using the load_signatures function
        cell_order = draw_obs[cluster_label].sort_values()
    else:
        # load the regulon_list from a file using the load_signatures function
        cell_order = data.obs[cluster_label].sort_values()
    celltypes = sorted(list(set(data.obs[cluster_label])))

    # Regulon specificity scores (RSS) across predicted cell types
    if rss_fn is None:
        rss_cellType = regulon_specificity_scores(auc_mtx, data.obs[cluster_label])
    else:
        rss_cellType = pd.read_csv(rss_fn, index_col=0)
    # Select the top 5 regulon_list from each cell type
    topreg = get_top_regulons_uneven(data, cluster_label, rss_cellType, topn=topn, target_celltype=target_celltype)

    if cluster_list is None:
        cluster_list = celltypes.copy()
    colorsd = dict((i, c) for i, c in zip(cluster_list, COLORS))
    colormap = [colorsd[x] for x in cell_order]
    # plot legend
    plot_legend(colorsd, fn=legend_fn)

    # plot z-score
    auc_zscore = cal_zscore(auc_mtx)
    plot_data = auc_zscore[topreg].loc[cell_order.index]
    sns.set(font_scale=1.2)
    g = sns.clustermap(plot_data,
                       annot=False,
                       square=False,
                       linecolor='gray',
                       yticklabels=True, xticklabels=True,
                       vmin=-3, vmax=3,
                       cmap='magma',  # "YlGnBu",
                       row_colors=colormap,
                       row_cluster=False, col_cluster=True,
                       figsize=(3 * CM, 5.5 * CM))
    g.cax.set_visible(True)
    g.ax_heatmap.set_yticks([])
    g.ax_heatmap.set_ylabel('')
    g.ax_heatmap.set_xlabel('')
    if save:
        file_format = os.path.splitext(fn)[1].replace('.','')
        plt.savefig(fn, format=file_format)
    return g


def auc_heatmap_reorder(data: anndata.AnnData,
                        auc_mtx: pd.DataFrame,
                        cluster_label: str,
                        rss_fn: str,
                        order_fn: Optional[str] = None,
                        topn=10,
                        save=True,
                        subset=True,
                        subset_size=5000,
                        fn='clusters_heatmap_top5.png',
                        legend_fn="rss_celltype_legend.png",
                        cluster_list=None,
                        **kwargs):
    """
    Plot heatmap for Regulon specificity scores (RSS) value
    :param data:
    :param auc_mtx:
    :param cluster_label:
    :param rss_fn:
    :param topn:
    :param save:
    :param subset:
    :param subset_size:
    :param fn:
    :param legend_fn:
    :param cluster_list: list of cluster names one prefer to use
    :return:

    Example:
        # only plot ['CNS', 'amnioserosa', 'carcass'] clusters and their corresponding top regulons
        auc_heatmap(adata, auc_mtx, cluster_label='celltypes', subset=False,
                    rss_fn='regulon_specificity_scores.txt',
                    cluster_list=['CNS', 'amnioserosa', 'carcass'])
    """
    # 1. Custom regulons
    # Select top regulons
    if order_fn is None:
        # Regulon specificity scores (RSS) across predicted cell types
        if rss_fn is None:
            rss_cellType = regulon_specificity_scores(auc_mtx, data.obs[cluster_label])
        else:
            rss_cellType = pd.read_csv(rss_fn, index_col=0)
        topreg = get_top_regulons(data, cluster_label, rss_cellType, topn=topn)
    else:
        with open(order_fn, 'r') as f:
            topreg = f.read().splitlines()

    # # 2. If Subset
    # if subset and len(data.obs) > subset_size:
    #     fraction = subset_size / len(data.obs)
    #     # do stratified sampling
    #     draw_obs = data.obs.groupby(cluster_label, group_keys=False).apply(lambda x: x.sample(frac=fraction))
    #     # load the regulon_list from a file using the load_signatures function
    #     cell_order = draw_obs[cluster_label].sort_values()
    # else:
    #     # load the regulon_list from a file using the load_signatures function
    #     cell_order = data.obs[cluster_label].sort_values()

    # #3. Custom cell type/cluster order
    # SET CELL ORDERS
    celltypes = sorted(list(set(data.obs[cluster_label])))
    if cluster_list is None:
        # cluster_list = celltypes
        cell_order = data.obs[cluster_label].sort_values()
    else:  # when provide cluster list
        # sort cell types in a custom order
        data.obs[cluster_label] = pd.Categorical(data.obs[cluster_label], categories=cluster_list, ordered=True)
        data.obs = data.obs.sort_values(cluster_label)
        cell_order = data.obs[cluster_label]

    # 4. Make plotting data
    # map color to cell type
    colorsd = dict((i, c) for i, c in zip(celltypes, COLORS))
    colormap = [colorsd[x] for x in cell_order]

    # plot legend
    plot_legend(colorsd, fn=legend_fn)

    # plot z-score
    auc_zscore = cal_zscore(auc_mtx)
    plot_data = auc_zscore[topreg].loc[cell_order.index]
    sns.set(font_scale=1.2)
    g = sns.clustermap(plot_data,
                       annot=False,
                       square=False,
                       linecolor='gray',
                       yticklabels=True, xticklabels=True,
                       vmin=-1.5, vmax=2.5,
                       cmap="YlGnBu",
                       row_colors=colormap,
                       row_cluster=False, col_cluster=True,
                       **kwargs)
    g.cax.set_visible(False)
    g.ax_heatmap.set_yticks([])
    g.ax_heatmap.set_ylabel('')
    g.ax_heatmap.set_xlabel('')
    if save:
        plt.tight_layout()
        file_format = os.path.splitext(fn)[1].replace('.','')
        plt.savefig(fn, format=file_format)
    return g


def generate_plot_data(data, auc_mtx, cluster_label, mode='mean', subset=False, subset_size=None, rss_fn=None, order_fn=None,
                       topn=None, cluster_list=None):
    if cluster_list:
        data = data[data.obs[cluster_label].isin(cluster_list)].copy()
        auc_mtx = auc_mtx.loc[list(data.obs_names)]

    if subset and len(data.obs) > subset_size:
        fraction = subset_size / len(data.obs)
        # do stratified sampling
        draw_obs = data.obs.groupby(cluster_label, group_keys=False).apply(lambda x: x.sample(frac=fraction))
        # load the regulon_list from a file using the load_signatures function
        cell_order = draw_obs[cluster_label].sort_values()
    else:
        # load the regulon_list from a file using the load_signatures function
        cell_order = data.obs[cluster_label].sort_values()

    celltypes = sorted(list(set(data.obs[cluster_label])))

    # Regulon specificity scores (RSS) across predicted cell types
    if rss_fn is None:
        rss_cellType = regulon_specificity_scores(auc_mtx, data.obs[cluster_label])
    else:
        rss_cellType = pd.read_csv(rss_fn, index_col=0)

    # load regulon order list
    if order_fn:
        with open(order_fn, 'r') as f:
            topreg = f.read().splitlines()
    elif topn is None:
        topreg = list(auc_mtx.columns)
    else:
        topreg = get_top_regulons(data, cluster_label, rss_cellType, topn=topn)

    if cluster_list is None:
        cluster_list = celltypes.copy()
    colorsd = dict(zip(cluster_list, COLORS))
    colormap = [colorsd[x] for x in cluster_list]

    # plot z-score
    auc_zscore = cal_zscore(auc_mtx)
    try:
        plot_data = auc_zscore[topreg].loc[cell_order.index]
    except KeyError:
        com_topreg = list(set(topreg).intersection(set(auc_zscore.columns)))
        plot_data = auc_zscore[com_topreg].loc[cell_order.index]
    # calculate mean values for each celltype
    plot_data['celltype'] = cell_order
    if mode == 'mean':
        plot_data = plot_data.groupby(['celltype']).mean()
    return plot_data, colorsd, colormap


def auc_heatmap_reorder2(data: anndata.AnnData,
                         auc_mtx: pd.DataFrame,
                         cluster_label: str,
                         rss_fn: str,
                         order_fn: Optional[str] = None,
                         topn: Optional[int] = 10,
                         save=True,
                         subset=True,
                         subset_size=5000,
                         fn='clusters_heatmap_top5.png',
                         legend_fn="rss_celltype_legend.png",
                         cluster_list=None,
                         figsize=(3 * 3 * CM, 5.5 * 3 * CM),
                         annot=False,
                         square=False,
                         linecolor='gray',
                         yticklabels=True,
                         xticklabels=5,
                         vmin=-1,
                         vmax=1,
                         cmap="magma",
                         row_cluster=False,
                         col_cluster=False,
                         **kwargs):
    """
    Plot heatmap for Regulon specificity scores (RSS) value
    :param data:
    :param auc_mtx:
    :param cluster_label:
    :param rss_fn:
    :param order_fn:
    :param topn:
    :param save:
    :param subset:
    :param subset_size:
    :param fn:
    :param legend_fn:
    :param cluster_list: list of cluster names one prefer to use
    :param figsize:
    :param annot:
    :param square:
    :param linecolor:
    :param yticklabels:
    :param xticklabels:
    :param vmin:
    :param vmax:
    :param cmap:
    :param row_cluster:
    :param col_cluster:
    :return:

    Example:
    # only plot ['CNS', 'amnioserosa', 'carcass'] clusters and their corresponding top regulons
    auc_heatmap(adata, auc_mtx, cluster_label='celltypes', subset=False,
                rss_fn='regulon_specificity_scores.txt',
                cluster_list=['CNS', 'amnioserosa', 'carcass'],
                xticklabels='auto')
    """
    plot_data, colorsd, colormap = generate_plot_data(data, auc_mtx, cluster_label, subset=subset,
                                                      subset_size=subset_size, rss_fn=rss_fn, order_fn=order_fn,
                                                      topn=topn, cluster_list=cluster_list)
    # generate legend file
    plot_legend(colorsd, fn=legend_fn)

    # sns.set(font_scale=1.2)
    g = sns.clustermap(plot_data,
                       annot=annot,
                       square=square,
                       linecolor=linecolor,
                       yticklabels=yticklabels, xticklabels=xticklabels,
                       vmin=vmin, vmax=vmax,
                       cmap=cmap,  # YlGnBu, RdYlBu
                       row_colors=colormap,
                       row_cluster=row_cluster, col_cluster=col_cluster,
                       figsize=figsize,
                       **kwargs)
    g.cax.set_visible(True)
    g.ax_heatmap.set_ylabel('')
    g.ax_heatmap.set_xlabel('')
    plt.setp(g.ax_heatmap.yaxis.get_majorticklabels(), rotation=0)
    if save:
        plt.tight_layout()
        file_format = os.path.splitext(fn)[1].replace('.','')
        plt.savefig(fn, format=file_format)
    return g


def auc_heatmap_reorder3(data: anndata.AnnData,
                         auc_mtx: pd.DataFrame,
                         cluster_label: str,
                         rss_fn: str,
                         order_fn: str,
                         target_celltype: str = 'ventricular-specific CM',
                         save=True,
                         subset=True,
                         subset_size=5000,
                         fn='clusters_heatmap_top5.png',
                         legend_fn="rss_celltype_legend.png",
                         cluster_list=None):
    """
    Plot heatmap for Regulon specificity scores (RSS) value
    :param data:
    :param auc_mtx:
    :param cluster_label:
    :param rss_fn:
    :param save:
    :param subset:
    :param subset_size:
    :param fn:
    :param legend_fn:
    :param cluster_list: list of cluster names one prefer to use
    :return:

    Example:
        # only plot ['CNS', 'amnioserosa', 'carcass'] clusters and their corresponding top regulons
        auc_heatmap(adata, auc_mtx, cluster_label='celltypes', subset=False,
                    rss_fn='regulon_specificity_scores.txt',
                    cluster_list=['CNS', 'amnioserosa', 'carcass'])
    """
    if subset and len(data.obs) > subset_size:
        fraction = subset_size / len(data.obs)
        # do stratified sampling
        draw_obs = data.obs.groupby(cluster_label, group_keys=False).apply(lambda x: x.sample(frac=fraction))
        # load the regulon_list from a file using the load_signatures function
        cell_order = draw_obs[cluster_label].sort_values()
    else:
        # load the regulon_list from a file using the load_signatures function
        cell_order = data.obs[cluster_label].sort_values()
    celltypes = sorted(list(set(data.obs[cluster_label])))

    # Regulon specificity scores (RSS) across predicted cell types
    if rss_fn is None:
        rss_cellType = regulon_specificity_scores(auc_mtx, data.obs[cluster_label])
    else:
        rss_cellType = pd.read_csv(rss_fn, index_col=0)

    # load regulon order list
    with open(order_fn, 'r') as f:
        topreg = f.read().splitlines()

    if cluster_list is None:
        cluster_list = celltypes.copy()
    colorsd = dict((i, c) for i, c in zip(cluster_list, COLORS))
    colormap = [colorsd[x] for x in cell_order]

    # plot z-score
    auc_zscore = cal_zscore(auc_mtx)
    plot_data = auc_zscore[topreg].loc[cell_order.index]
    # calculate mean values for each celltype
    plot_data['celltype'] = cell_order

    plot_data = plot_data.groupby(["celltype"], sort=False).apply(
        lambda x: x.sort_values(list(plot_data.columns), ascending=False))  # .reset_index(drop=True)

    colormap = [colorsd[x] for x in plot_data.celltype]

    sns.set(font_scale=1.2)
    g = sns.clustermap(plot_data.drop(['celltype'], axis=1),
                       annot=False,
                       square=False,
                       linecolor='gray',
                       yticklabels=True, xticklabels=True,
                       vmin=-3, vmax=3,
                       cmap="YlGnBu",
                       row_colors=colormap,
                       row_cluster=False, col_cluster=False)  # YlGnBu, RdYlBu
    g.cax.set_visible(True)
    g.ax_heatmap.set_yticks([])
    g.ax_heatmap.set_ylabel('')
    g.ax_heatmap.set_xlabel('')
    if save:
        plt.tight_layout()
        file_format = os.path.splitext(fn)[1].replace('.','')
        plt.savefig(fn, format=file_format)
    return g


def func(df):
    total_cell_order = []
    ct = df['celltype']
    celltypes = sorted(set(ct))
    for celltype in celltypes:
        ct_df = df[df.celltype == celltype]
        g = sns.clustermap(ct_df.drop(['celltype'], axis=1),
                           annot=False,
                           square=False,
                           linecolor='gray',
                           yticklabels=True, xticklabels=True,
                           vmin=-1.5, vmax=2.5,
                           cmap="YlGnBu",
                           row_cluster=False, col_cluster=False)  # YlGnBu, RdYlBu
        labels = get_labels(g)
        total_cell_order += labels
        return total_cell_order


def get_labels(clustermap):
    labels = clustermap.ax_heatmap.yaxis.get_majorticklabels()
    cluster_labels = [label.get_text() for label in labels]
    return cluster_labels


def highlight_key(color_dir: dict,
                  new_value: str = '#8a8787',
                  key_to_highlight=None
                  ) -> dict:
    """
    Highlight one or more interested keys/celltypes when plotting,
    the rest of keys/celltypes will be set to gray by default.
    :param color_dir
    :param new_value
    :param key_to_highlight
    :return dict
    """
    # assert key_to_highlight in color_dir.keys()
    if key_to_highlight is None:
        key_to_highlight = ['Cardiac muscle lineages']
    for k, v in color_dir.items():
        if k not in key_to_highlight:
            color_dir[k] = new_value
    return color_dir


def plot_legend(color_dir, marker='o', linestyle='', numpoints=1, ncol=3, loc='center', figsize=(10, 5),
                fn='legend.pdf', **kwargs):
    """
    Make separate legend file for heatmap
    :param color_dir:
    :param linestyle: legend icon style
    :param marker: legend icon style
    :param numpoints:
    :param ncol: number of columns, legend layout
    :param loc: location of the legend
    :param figsize: (width, height)
    :param fn:
    :param kwargs:
    :return:

    Example:
        color_dir = {'celltype1': #000000}
        plot_legend(color_dir)
    """
    plt.figure(figsize=figsize)
    markers = [plt.Line2D([0, 0], [0, 0], color=color, marker=marker, linestyle=linestyle, **kwargs)
               for color in color_dir.values()]
    plt.legend(markers, color_dir.keys(), numpoints=numpoints, ncol=ncol, loc=loc, frameon=False)
    plt.box(False)
    plt.xticks([])
    plt.yticks([])
    plt.tight_layout()
    file_format = os.path.splitext(fn)[1].replace('.','')
    plt.savefig(fn, format=file_format)
    plt.close()


def get_top_regulons(data: anndata.AnnData,
                     cluster_label: str,
                     rss_celltype: pd.DataFrame,
                     topn: int) -> list:
    """
    get top n regulons for each cell type based on regulon specificity scores (rss)
    :param data:
    :param cluster_label:
    :param rss_celltype:
    :param topn:
    :return: a list
    """
    # Select the top 5 regulon_list from each cell type
    cats = sorted(list(set(data.obs[cluster_label])))
    # all cell type labels should be strings
    if not all(isinstance(x, str) for x in cats):
        cats = [str(x) for x in cats]
    rss_celltype.index = rss_celltype.index.astype(str)

    topreg = []
    for i, c in enumerate(cats):
        topreg.extend(
            list(rss_celltype.T[c].sort_values(ascending=False)[:topn].index)
        )
    topreg = list(set(topreg))
    return topreg


def get_top_regulons_uneven(data: anndata.AnnData,
                            cluster_label: str,
                            rss_celltype: pd.DataFrame,
                            topn: int = 10,
                            target_celltype: str = 'ventricular-specific CM',
                            target_topn: int = 20) -> list:
    """
    get target_topn regulons for interested cell type and topn regulons for the rest cell types
    based on regulon specificity scores (rss)
    :param data:
    :param cluster_label:
    :param rss_celltype:
    :param topn:
    :param target_celltype:
    :param target_topn:
    :return: a list
    """
    cats = sorted(list(set(data.obs[cluster_label])))
    topreg = []
    for i, c in enumerate(cats):
        if c == target_celltype:
            topreg.extend(list(rss_celltype.T[c].sort_values(ascending=False)[:target_topn].index))
        else:
            topreg.extend(list(rss_celltype.T[c].sort_values(ascending=False)[:topn].index))
    topreg = list(set(topreg))
    return topreg


def cal_zscore(auc_mtx: pd.DataFrame, save=False) -> pd.DataFrame:
    """
    calculate z-score for each gene among cells
    :param auc_mtx:
    :param save:
    :return:
    """
    func = lambda x: (x - x.mean()) / x.std(ddof=0)
    auc_zscore = auc_mtx.transform(func, axis=0)
    if save:
        auc_zscore.to_csv('auc_zscore.csv', index=False)
    return auc_zscore


def go_bar(fn):
    name = fn.split('.')[0]
    # df = pd.read_excel(fn)
    df = pd.read_csv(fn)
    CM = 1 / 2.54
    y_pos = range(len(df.Description))
    plt.figure(figsize=(20 * CM, 12 * CM))
    ax = plt.barh(y_pos, df['LogP'])
    plt.yticks(y_pos, labels=df.Description)
    plt.tick_params(axis='y', length=0)
    # ax.yaxis.tick_right()
    # plt.yticks([])
    plt.tight_layout()
    plt.savefig(f'{name}.pdf', format='pdf')
    plt.close()


def plot_3D_legend(adata, cluster_label='leiden', prefix=''):
    """
    :param adata:
    :param cluster_label:
    :param is_deconv:
    :param prefix:
    :return:

    Example:
        plot_3D_legend(adata, cluster_label='leiden', prefix=n)
    """
    import plotly.express as px

    obs = adata.obs.copy()
    obs['x'] = adata.obsm['spatial'][:, 0]
    obs['y'] = adata.obsm['spatial'][:, 1]
    obs['z'] = adata.obsm['spatial'][:, 2]
    xs = adata.obsm['spatial'][:, 0]
    ys = adata.obsm['spatial'][:, 1]
    zs = adata.obsm['spatial'][:, 2]
    fig = px.scatter_3d(obs,
                        x=xs,
                        y=ys,
                        z=zs,
                        color=cluster_label,
                        opacity=0.7,
                        )
    fig.update_traces(marker_size=3)
    fig.layout.update(
        scene=dict(aspectmode='data')
    )
    fig.write_html(f'{prefix}_{cluster_label}.html')


def spatial_plot_2d(adata, color='annotation', prefix='2d_plot'):
    """

    :param adata:
    :param color:
    :param prefix:
    :return:

    Example:
        spatial_plot_2d(adata, color='leiden', prefix=prefix)
    """
    import matplotlib
    palette = list(matplotlib.colors.CSS4_COLORS.values())

    x, y = zip(*adata.obsm['spatial'])
    annotations = adata.obs[color].unique()
    colors = palette[:len(annotations)]
    color_dict = dict((anno, color) for anno, color in zip(annotations, colors))
    c = [color_dict[anno] for anno in adata.obs[color]]
    plt.scatter(x, y, s=1, c=c, lw=0, edgecolors='none')
    plt.gca().set_aspect('equal')
    plt.show()
    plt.savefig(f'{prefix}_{color}_2d.png')
    plt.close()


def plot_2d(data: anndata.AnnData,
            auc_mtx,
            reg_name: str,
            fn: str,
            pos_label='spatial',
            marker='.',
            edgecolors='none',
            cmap='plasma',
            lw=0,
            **kwargs):
    """
    Plot genes of one regulon on a 2D map
    :param pos_label:
    :param data:
    :param auc_mtx:
    :param reg_name:
    :param fn:
    :return:
    """
    if '(+)' not in reg_name:
        reg_name = reg_name + '(+)'

    if isinstance(data.obsm[pos_label], np.ndarray):
        cell_coor = data.obsm[pos_label]
    else:
        cell_coor = data.obsm[pos_label].to_numpy()
    auc_zscore = cal_zscore(auc_mtx)
    # prepare plotting data
    sub_zscore = auc_zscore[reg_name]
    # sort data points by zscore (low to high), because first dot will be covered by latter dots
    zorder = np.argsort(sub_zscore.values)
    # plot cell/bin dot, x y coor
    sc = plt.scatter(cell_coor[:, 0][zorder],
                     cell_coor[:, 1][zorder],
                     c=sub_zscore.iloc[zorder],
                     marker=marker,
                     edgecolors=edgecolors,
                     cmap=cmap,
                     lw=lw,
                     **kwargs)
    plt.axis("equal")
    plt.title(reg_name)
    plt.box(False)
    plt.axis('off')
    plt.colorbar(sc, shrink=0.35)
    file_format = os.path.splitext(fn)[1].replace('.', '')
    plt.savefig(fn, format=file_format)
    plt.close()


def plot_celltype(adata, color='annotation', fn='cell_type.png', custom_labels=None, spatial_label='spatial', s=1, marker='.'):
    """

    :param adata:
    :param color:
    :param prefix:
    :param custom_labels: labels associate with cell types. e.g. it's celltype1 in annotation, but you want to label it as 'ct1'
    :param spatial_label:
    :return:

    Example:
        tfs = ['Adf1', 'Aef1', 'grh', 'kn', 'tll']
        plot_celltype(adata, color='celltype', prefix='ground_truth', custom_labels=tfs)
    """
    annotations = adata.obs[color].unique()
    import matplotlib.cm as cm
    colors = cm.rainbow(np.linspace(0, 1, len(annotations)))
    color_dict = dict((anno, color) for anno, color in zip(annotations, colors))

    if custom_labels:
        labels = custom_labels
    else:
        labels = list(annotations)

    for i, label in enumerate(labels):
        # data = adata[adata.obs[color] == (i + 1)]
        data = adata[adata.obs[color] == label]
        c = [color_dict[anno] for anno in data.obs[color]]
        # plt.scatter(data.obsm[spatial_label]['x'], data.obsm[spatial_label]['y'], s=s, c=c, marker=marker, label=label)
        plt.scatter(data.obsm[spatial_label][:, 0], data.obsm[spatial_label][:, 1], s=s, c=c, marker=marker, label=label)
    plt.gca().set_aspect('equal')
    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    plt.tight_layout()
    if fn is None:
        fn ='cell_type.png'
    file_format = os.path.splitext(fn)[1].replace('.', '')
    plt.savefig(fn, format=file_format)
    plt.close()


def plot_leiden(adata):
    colors = adata.uns['leiden_colors']
    celltypes = list(adata.obs['leiden'].unique())
    color_dict = dict(zip(celltypes, colors))
    for i, label in enumerate(celltypes):
        data = adata[adata.obs['leiden'] == label]
        c = [color_dict[anno] for anno in data.obs['leiden']]
        plt.scatter(data.obsm['spatial']['x'], data.obsm['spatial']['y'], s=1, c=c, marker='.', label=label)
    plt.gca().set_aspect('equal')
    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    plt.tight_layout()
    plt.box(False)
    plt.axis('off')
    plt.savefig('mouse_brain_E15_leiden_spatial_0.9.png')
    plt.show()
    plt.close()


def plot_cancer(adata):
    colors = COLORS
    celltypes = list(adata.obs['bayes_clusters'].unique())
    color_dict = dict(zip(celltypes, colors))
    for i, label in enumerate(celltypes):
        data = adata[adata.obs['bayes_clusters'] == label]
        c = [color_dict[anno] for anno in data.obs['bayes_clusters']]
        plt.scatter(data.obsm['spatial']['x'], data.obsm['spatial']['y'], s=1, c=c, marker='.', label=label)
    plt.gca().set_aspect('equal')
    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    plt.tight_layout()
    plt.box(False)
    plt.axis('off')
    plt.savefig('cancer_celltype.pdf', format='pdf')
    plt.close()


def plot_ct(adata, colors, ctlabel='leiden'):
    celltypes = list(adata.obs[ctlabel].unique())
    color_dict = dict(zip(celltypes, colors))
    for i, label in enumerate(celltypes):
        data = adata[adata.obs[ctlabel] == label]
        c = [color_dict[anno] for anno in data.obs[ctlabel]]
        plt.scatter(data.obsm['spatial']['x'], data.obsm['spatial']['y'], s=50, c=c, marker='.', label=label)
    plt.gca().set_aspect('equal')
    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    plt.tight_layout()
    plt.show()


def plot_gene(data: anndata.AnnData,
              gene_name: str,
              fn: str,
              pos_label='spatial',
              show_bg=False,
              marker='.',
              edgecolors='none',
              cmap='plasma',
              lw=0,
              **kwargs):
    """
    Plot a gene on a 3D map
    :param lw:
    :param cmap:
    :param edgecolors:
    :param marker:
    :param show_bg:
    :param pos_label:
    :param data:
    :param gene_name:
    :param fn:
    :return:

    Example:
        plot_gene(data, 'spatial', 'Zfp354c', 'Zfp354c.png')
    """
    # prepare plotting data
    cell_coor = data.obsm[pos_label]
    exp_mtx = data.to_df()

    sc = plt.scatter(cell_coor['x'],
                     cell_coor['y'],
                     c=exp_mtx[gene_name],
                     marker=marker,
                     edgecolors=edgecolors,
                     cmap=cmap,
                     lw=lw, **kwargs)

    if not show_bg:
        plt.box(False)
        plt.axis('off')
    plt.axis("equal")
    plt.title(gene_name)
    plt.colorbar(sc, shrink=0.35)
    file_format = os.path.splitext(fn)[1].replace('.', '')
    plt.savefig(fn, format=file_format)
    plt.close()


def plot_ligand_receptor(data,
                         ligand: str,
                         receptor: str,
                         fn: str,
                         pos_label='spatial',
                         show_bg=False):
    """

    :param data:
    :param ligand:
    :param receptor:
    :param fn:
    :param pos_label:
    :param show_bg:
    :return:
    """
    # prepare plotting data
    cell_coor = data.obsm[pos_label]
    exp_mtx = data.to_df()

    # coor when exp is more than 0
    x = cell_coor['x']
    y = cell_coor['y']
    data1 = exp_mtx[ligand]
    data2 = exp_mtx[receptor]
    z1 = cal_zscore(exp_mtx[ligand])
    z2 = cal_zscore(exp_mtx[receptor])

    plt.scatter(x,
                y,
                c=data1,
                marker='.',
                edgecolors='none',
                cmap='Blues',
                lw=0, label='ligand', alpha=0.5)
    plt.scatter(x,
                y,
                c=data2,
                marker='.',
                edgecolors='none',
                cmap='Oranges',
                lw=0, label='receptor', alpha=0.5)

    # Highlight overlapping points with a different color
    for i in range(len(x)):
        if z1[i] > 1 and z2[i] > 1:
            plt.scatter(x[i], y[i], color='red', marker='.', s=100)

    if not show_bg:
        plt.box(False)
        plt.axis('off')
    plt.axis("equal")
    plt.title(f'{ligand}_{receptor}')
    # plt.colorbar(sc, shrink=0.35)
    file_format = os.path.splitext(fn)[1].replace('.', '')
    plt.savefig(fn, format=file_format)
    plt.close()


class PlotDataParameters:
    def __init__(self, cluster_list=None, cluster_label=None, subset=None, subset_size=None, rss_fn=None, order_fn=None, topn=None, mode=None):
        self.cluster_list = cluster_list
        self.cluster_label = cluster_label
        self.subset = subset
        self.rss_fn = rss_fn
        self.order_fn = order_fn
        self.topn = topn
        self.subset_size = subset_size
        self.mode = mode


def generate_plot_data(data, auc_mtx, parameters):
    cluster_list = parameters.cluster_list
    subset = parameters.subset
    rss_fn = parameters.rss_fn
    order_fn = parameters.order_fn
    topn = parameters.topreg
    subset_size = parameters.subset_size
    mode = parameters.mode
    cluster_label = parameters.cluster_label

    # Generate plot data based on the provided parameters
    if cluster_list:
        data = data[data.obs[cluster_label].isin(cluster_list)].copy()
        auc_mtx = auc_mtx.loc[list(data.obs_names)]

    if subset and len(data.obs) > subset_size:
        fraction = subset_size / len(data.obs)
        # do stratified sampling
        draw_obs = data.obs.groupby(cluster_label, group_keys=False).apply(lambda x: x.sample(frac=fraction))
        # load the regulon_list from a file using the load_signatures function
        cell_order = draw_obs[cluster_label].sort_values()
    else:
        # load the regulon_list from a file using the load_signatures function
        cell_order = data.obs[cluster_label].sort_values()

    celltypes = sorted(list(set(data.obs[cluster_label])))

    # Regulon specificity scores (RSS) across predicted cell types
    if rss_fn is None:
        rss_cellType = regulon_specificity_scores(auc_mtx, data.obs[cluster_label])
    else:
        rss_cellType = pd.read_csv(rss_fn, index_col=0)

    # load regulon order list
    if order_fn:
        with open(order_fn, 'r') as f:
            topreg = f.read().splitlines()
    elif topn is None:
        topreg = list(auc_mtx.columns)
    else:
        topreg = get_top_regulons(data, cluster_label, rss_cellType, topn=topn)

    if cluster_list is None:
        cluster_list = celltypes.copy()
    colorsd = dict(zip(cluster_list, COLORS))
    colormap = [colorsd[x] for x in cluster_list]

    # plot z-score
    auc_zscore = cal_zscore(auc_mtx)
    try:
        plot_data = auc_zscore[topreg].loc[cell_order.index]
    except KeyError:
        com_topreg = list(set(topreg).intersection(set(auc_zscore.columns)))
        plot_data = auc_zscore[com_topreg].loc[cell_order.index]
    # calculate mean values for each celltype
    plot_data['celltype'] = cell_order
    if mode == 'mean':
        plot_data = plot_data.groupby(['celltype']).mean()
    return plot_data, colorsd, colormap


def t():
    exp = np.array()
    array = np.array()
    for i,row in coordinates:
        x,y = row
        array[x,y] = exp[i,:]


def plot_isr(mtx,
             cell_coor,
             reg_name: str,
             receptor_name,
             fn: str,
             show_bg=False,
             marker='.',
             edgecolors='none',
             cmap='plasma',
             lw=0,
             s=1,
             **kwargs):
    sc = plt.scatter(cell_coor['x'],
                     cell_coor['y'],
                     c=mtx,
                     marker=marker,
                     edgecolors=edgecolors,
                     cmap=cmap,
                     s=s,
                     lw=lw, **kwargs)
    if not show_bg:
        plt.box(False)
        plt.axis('off')
    plt.axis("equal")
    plt.title(f'{reg_name}(+) {receptor_name}')
    plt.colorbar(sc, shrink=0.35)
    file_format = os.path.splitext(fn)[1].replace('.', '')
    plt.savefig(fn, format=file_format)
    plt.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@file: spagrn_parser.py
@time: 2023/Nov/01
@description: test file for inference gene regulatory networks module
@author: Yao LI
@email: liyao1@genomics.cn
@last modified by: Yao LI
"""

import os
import sys
import argparse
from multiprocessing import cpu_count
from spagrn.regulatory_network import InferNetwork as irn
import spagrn.plot as prn


# def receptor_command(args):
#     data = irn.read_file(args.expression_mtx_fname)
#     grn = irn(data)
#     grn.get_receptors(args.reference, receptor_key=args.receptor_key, save_tmp=args.save_tmp)
#     data.write_h5ad(args.expression_mtx_fname)


def inference_command(args):
    # avoid TypeError: stat: path should be string, bytes, os.PathLike or integer, not _io.TextIOWrapper
    out_dir = args.output
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    prefix = os.path.join(out_dir, args.method)
    # load data
    data = irn.read_file(args.expression_mtx_fname)
    data = irn.preprocess(data)
    # create grn
    grn = irn(data)
    # run_all analysis
    grn.infer(args.database,
              args.motif,
              args.tfs_fname,
              niche_df=args.reference,
              num_workers=args.num_workers,
              cache=args.cache,
              save_tmp=args.save_tmp,
              c_threshold=args.c_threshold,
              layers=args.layer_key,
              latent_obsm_key=args.latent_obsm_key,
              model=args.model,
              n_neighbors=args.n_neighbors,
              weighted_graph=args.weighted_graph,
              cluster_label=args.cluster_label,
              method=args.method,
              noweights=args.noweights,
              rho_mask_dropouts=args.rho_mask_dropouts,
              project_name=prefix)


def add_computation_parameters(parser):
    group = parser.add_argument_group("computation arguments")
    group.add_argument(
        "--num_workers",
        type=int,
        default=cpu_count(),
        help="",
    )
    group.add_argument(
        "--save_tmp",
        default=True,
        help="",
    )
    group.add_argument(
        "--cache",
        default=False,
        help="",
    )
    return parser


def add_coexp_parameters(parser):
    group = parser.add_argument_group("co-expressed module generation arguments")
    # SPG params
    group.add_argument(
        "--c_threshold",
        type=float,
        default=-1,
        help="",
    )
    group.add_argument(
        "--layer_key",
        type=str,
        default=None,
        help="",
    )
    group.add_argument(
        "--model",
        type=str,
        default='danb',
        choices=["bernoulli", "danb", "normal", "none"],
        help="",
    )
    group.add_argument(
        "--latent_obsm_key",
        type=str,
        default='spatial',
        help="",
    )
    group.add_argument(
        "--n_neighbors",
        type=int,
        default=30,
        help="",
    )
    group.add_argument(
        "--fdr_threshold",
        type=float,
        default=0.05,
        help="",
    )
    group.add_argument(
        "--distances_obsp_key",
        type=str,
        help="",
    )
    group.add_argument(
        "--neighborhood_factor",
        type=str,
        help="",
    )
    group.add_argument(
        "--approx_neighbors",
        type=str,
        help="",
    )
    group.add_argument(
        "--weighted_graph",
        default=False,
        help="",
    )
    # SCC params
    group.add_argument(
        "--sigma",
        type=int,
        default=15,
        help="",
    )
    group.add_argument(
        "--zero_cutoff",
        type=int,
        default=5,
        help="",
    )
    group.add_argument(
        "--cor_method",
        type=str,
        default='spearman',
        help="",
    )
    return parser


def add_ctx_parameters(parser):
    group = parser.add_argument_group("Regulons Generation arguments")
    group.add_argument(
        "--rho_mask_dropouts",
        action="store_const",
        const=True,
        default=False,
        help="",
    )
    group.add_argument(
        "--rank_threshold",
        type=int,
        default=1500,
        help="",
    )
    group.add_argument(
        "--prune_auc_threshold",
        type=float,
        default=0.5,
        help="",
    )
    group.add_argument(
        "--nes_threshold",
        type=float,
        default=0.5,
        help="",
    )
    group.add_argument(
        "--motif_similarity_fdr",
        type=float,
        default=0.05,
        help="",
    )
    group.add_argument(
        "--orthologuous_identity_threshold",
        help="",
    )
    group.add_argument(
        "--weighted_recovery",
        help="",
    )
    group.add_argument(
        "--module_chunksize",
        help="",
    )


def add_aucell_parameters(parser):
    group = parser.add_argument_group("Regulon Activity Calculation arguments")
    group.add_argument(
        "--auc_threshold",
        type=float,
        default=0.5,
        help="",
    )
    group.add_argument(
        "--noweights",
        default=False,
        help="",
    )
    group.add_argument(
        "--normalize",
        default=False,
        help="",
    )
    group.add_argument(
        "--seed",
        type=int,
        required=False,
        default=None,
        help="seed for generating random numbers",
    )


def add_receptor_parameters(parser):
    group = parser.add_argument_group("Receptors Detection arguments")
    group.add_argument(
        "--reference",
        # type=argparse.FileType("r"),
        type=str,
        help="",
        default=None,
        required=False,
    )
    group.add_argument(
        "--receptor_key",
        type=str,
        default='to',
        help="",
        required=False,
    )


def create_argument_parser():
    parser = argparse.ArgumentParser(
        prog='spagrn',
        # prog=os.path.splitext(os.path.basename(__file__))[0],
        description="Spatial Gene Regulatory Network inference",
        fromfile_prefix_chars="@",
        add_help=True,
        epilog="Arguments can be read from file using a @args.txt construct. "
    )

    subparsers = parser.add_subparsers(help="sub-command help")

    # --------------------------------------------
    # create the parser for the "infer" command
    # --------------------------------------------
    parser_infer = subparsers.add_parser(
        "infer", help="Derive regulons from SRT expression matrix and spatial coordinates."
    )
    parser_infer.add_argument(
        "expression_mtx_fname",
        # type=argparse.FileType("r"),
        type=str,
        help="The name of the file that contains the expression matrix for the single cell experiment."
             " Two file formats are supported: csv (rows=cells x columns=genes) or loom (rows=genes x columns=cells).",
    )
    parser_infer.add_argument(
        "tfs_fname",
        # type=argparse.FileType("r"),
        type=str,
        help="TF list file.",
    )
    parser_infer.add_argument(
        "-db",
        "--database",
        # type=argparse.FileType("r"),
        type=str,
        required=True,
        help="ranked motifs database file, in feather format.",
    )
    parser_infer.add_argument(
        "--motif",
        # type=argparse.FileType("r"),
        type=str,
        required=True,
        help="motif annotation file, in tbl format.",
    )
    parser_infer.add_argument(
        "-m",
        "--method",
        choices=["spg", "scc"],
        default="spg",
        help="The algorithm for gene regulatory network reconstruction (default: spg).",
    )
    parser_infer.add_argument(
        "-c",
        "--cluster_label",
        default="annotation",
        help="label storing cell type/cluster annotation.",
    )
    parser_infer.add_argument(
        "-o",
        "--output",
        # type=argparse.FileType("w"),
        type=str,
        # default=str(sys.stdout),
        help="Output file/stream, i.e. the adjacencies table with correlations (csv, tsv).",
    )
    add_computation_parameters(parser_infer)
    add_coexp_parameters(parser_infer)
    add_ctx_parameters(parser_infer)
    add_aucell_parameters(parser_infer)
    add_receptor_parameters(parser_infer)
    parser_infer.set_defaults(func=inference_command)

    # ----------------------------------------------
    # create the parser for the "receptor" command
    # ----------------------------------------------
    # parser_receptor = subparsers.add_parser(
    #     "receptor",
    #     help="[Optional] Receptor detection.",
    # )
    # parser_receptor.add_argument(
    #     "spagrn_output_fname",
    #     type=str,
    #     help="",
    # )
    # add_receptor_parameters(parser_receptor)
    # parser_receptor.set_defaults(func=receptor_command)

    # ------------------------------------------
    # create the parser for the "plot" command
    # ------------------------------------------
    # 'plot' subcommand
    plot_parser = subparsers.add_parser('plot', help='Plot subcommand')
    plot_parser.add_argument(
        '-d',
        '--data',
        type=str,
        # required=True,
        help='Path to the data file'
    )
    plot_parser.add_argument(
        '-n',
        '--name',
        # required=True,
        type=str,
        help='name for gene/regulon to plot.'
    )
    plot_parser.add_argument(
        '--color',
        default='celltype',
        type=str,
        help=''
    )
    plot_parser.add_argument(
        "-m",
        "--mode",
        type=str,
        choices=['gene', 'regulon', 'celltype'],
        default='regulon',
        help="choose to plot gene, regulon or celltype. (default: regulon) ",
    )
    plot_parser.add_argument(
        "--dimension",
        type=str,
        choices=['2d', '3d', '2D', '3D'],
        default='2d',
        help="choose data dimension to plot, choices are {2d, 3d, 2D, 3D}. (default: 2d) ",
    )
    plot_parser.add_argument(
        '-p',
        '--pos_label',
        default='spatial',
        type=str,
        help=''
    )
    plot_parser.add_argument(
        '--custom_labels',
        default=None,
        help=''
    )
    plot_parser.add_argument(
        '-o',
        '--output',
        # required=True,
        type=str,
        help=''
    )
    add_plot_parameters(plot_parser)
    add_3d_parameters(plot_parser)
    plot_parser.set_defaults(func=plot_command)

    # for more plotting functions/alternatives
    # make subcommands
    plot_subparsers = plot_parser.add_subparsers(title='plot subcommands', dest='plot_command',
                                                 help='plot subcommand help')

    # 'heatmap' subcommand under 'plot'
    heatmap_parser = plot_subparsers.add_parser('heatmap', help='Generate a heatmap')
    heatmap_parser.add_argument(
        '-d',
        '--data',
        type=str,
        required=True,
        help='Path to the data file'
    )
    heatmap_parser.add_argument(
        '--cluster_label',
        type=str,
        default='annotation',
        help='label for cell type/cluster annotation'
    )
    heatmap_parser.add_argument(
        '--rss_fn',
        type=str,
        default=None,
        help='Regulon specificity scores (RSS) file, a text file.'
    )
    heatmap_parser.add_argument(
        '--topn',
        type=int,
        default=10,
        help='select top N regulons for each cell type'
    )
    heatmap_parser.add_argument(
        '--subset',
        default=False,
        help='If subset data by cells'
    )
    heatmap_parser.add_argument(
        '--subset_size',
        type=int,
        default=5000,
        help='If subset, number of cells to keep.'
    )
    heatmap_parser.add_argument(
        '-o',
        '--output',
        type=str,
        default='clusters_heatmap_top.png',
        required=True,
        help='Output image file name.'
    )
    heatmap_parser.add_argument(
        '--legend_fn',
        type=str,
        default='rss_celltype_legend.png',
        help='Output legend image file name.'
    )
    heatmap_parser.add_argument(
        '--cluster_list',
        default=None,
        help='list of clusters to map celltype-color, generate legend.'
    )
    heatmap_parser.add_argument(
        '--row_cluster',
        default=False,
    )
    heatmap_parser.add_argument(
        '--col_cluster',
        default=False,
    )
    add_plot_parameters(heatmap_parser)
    heatmap_parser.set_defaults(func=heatmap_command)

    # 'heatmap' subcommand under 'plot'
    web_parser = plot_subparsers.add_parser('web', help='Generate plot html file via Plotly')
    web_parser.add_argument(
        '-d',
        '--data',
        type=str,
        required=True,
        help='Path to the data file'
    )

    return parser


def add_plot_parameters(parser):
    group = parser.add_argument_group("General Plots arguments")
    group.add_argument(
        '--figsize',
        default=(3, 5.5),
        help='heatmap figure size, a tuple. (default: (3, 5.5))'
    )
    group.add_argument(
        '--marker',
        type=str,
        default='.',
        help=''
    )
    group.add_argument(
        '--show_bg',
        default=False,
        help=''
    )
    group.add_argument(
        '--cmap',
        type=str,
        default="YlGnBu",
        help='Colormap for the heatmap'
    )
    group.add_argument(
        '--annot',
        default=False,
        help=''
    )
    group.add_argument(
        '--square',
        default=False,
        help=''
    )
    group.add_argument(
        '--linecolor',
        default="gray",
        help=''
    )
    group.add_argument(
        '--yticklabels',
        default=False,
        help=''
    )
    group.add_argument(
        '--xticklabels',
        default=True,
        help=''
    )
    group.add_argument(
        '--vmin',
        default=-3,
        help='minimum value'
    )
    group.add_argument(
        '--vmax',
        default=3,
        help='max value'
    )
    group.add_argument(
        '--s',
        default=1,
        type=float,
        help='scatter s'
    )
    group.add_argument(
        '--edgecolors',
        type=str,
        default='none',
        help=''
    )
    group.add_argument(
        '--lw',
        default=0,
        help='line weight'
    )


def add_3d_parameters(parser):
    group = parser.add_argument_group("3D Plots arguments")
    group.add_argument(
        '-vv',
        '--view_vertical',
        type=int,
        default=222,
        help='vertical angle to view to the 3D object. range from 0 to 360.'
    )
    group.add_argument(
        '-vh',
        '--view_horizontal',
        type=int,
        default=-80,
        help='horizontal angle to view the 3D object. range from 0 to 360.'
    )
    group.add_argument(
        '--xscale',
        type=float,
        default=1,
        help='scale of x axis. (default: 1)'
    )
    group.add_argument(
        '--yscale',
        type=float,
        default=1,
        help=''
    )
    group.add_argument(
        '--zscale',
        type=float,
        default=1,
        help=''
    )


def plot_command(args):
    import scanpy as sc
    adata = sc.read_h5ad(args.data)
    auc_mtx = adata.obsm['auc_mtx']
    if args.mode == 'celltype':
        prn.plot_celltype(adata, color=args.color, spatial_label=args.pos_label, custom_labels=args.custom_labels,
                          fn=args.output)
                          # marker=args.marker, s=args.s)  # why the program keep crushing after added this line
    elif args.mode == 'gene':
        prn.plot_gene(adata, gene_name=args.name, fn=args.output, pos_label=args.pos_label, show_bg=args.show_bg,
                      cmap=args.cmap)
                # marker=args.marker, edgecolors=args.edgecolors, lw=args.lw)
    elif args.mode == 'regulon':
        if args.dimension in ['2d', '2D']:
            prn.plot_2d(adata, auc_mtx, reg_name=args.name, fn=args.output, pos_label=args.pos_label, cmap=args.cmap)
                        # marker=args.marker, edgecolors=args.edgecolors, lw=args.lw)
        elif args.dimension in ['3d', '3D']:
            prn.plot_3d_reg(adata, auc_mtx, reg_name=args.name, fn=args.output, view_vertical=args.view_vertical,
                            view_horizontal=args.view_horizontal, show_bg=args.show_bg, pos_label=args.pos_label,
                            xscale=args.xscale, yscale=args.yscale, zscale=args.zscale)


def heatmap_command(args):
    import scanpy as sc
    adata = sc.read_h5ad(args.data)
    auc_mtx = adata.obsm['auc_mtx']
    prn.auc_heatmap(adata, auc_mtx, args.cluster_label, args.rss_fn, topn=args.topn, cmap=args.cmap,
                    subset=args.subset, subset_size=args.subset_size, fn=args.output, legend_fn=args.legend_fn,
                    cluster_list=args.cluster_list, row_cluster=args.row_cluster, col_cluster=args.col_cluster)


def main(argv=None):
    # Parse arguments.
    parser = create_argument_parser()
    args = parser.parse_args(args=argv)
    if not hasattr(args, "func"):
        parser.print_help()
    else:
        args.func(args)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date: Created on 30 Jun 2024 13:39
# @Author: Yao LI
# @File: SpaGRN/autocor.py
import os
import sys
import time

import scipy
import scanpy as sc
import numpy as np
import pandas as pd
import anndata as ad

from pynndescent import NNDescent
from scipy.stats import chi2, norm
from scipy import stats
from scipy.spatial.distance import pdist, squareform
from scipy.sparse import csr_matrix, issparse

from esda.getisord import G
from esda.moran import Moran
from esda.geary import Geary

import multiprocessing
from tqdm import tqdm


# -----------------------------------------------------#
# spatial weights
# -----------------------------------------------------#
def save_array(array, fn='array.json'):
    import json
    from json import JSONEncoder
    class NumpyEncoder(json.JSONEncoder):
        """ Special json encoder for numpy types """

        def default(self, obj):
            if isinstance(obj, np.integer):
                return int(obj)
            elif isinstance(obj, np.floating):
                return float(obj)
            elif isinstance(obj, np.ndarray):
                return obj.tolist()
            return json.JSONEncoder.default(self, obj)

    encodedNumpyData = json.dumps(array, cls=NumpyEncoder)  # use dump() to write array into file
    with open(fn, 'w', encoding='utf-8') as f:
        json.dump(encodedNumpyData, f)


def save_list(l, fn='list.txt'):
    with open(fn, 'w') as f:
        f.write('\n'.join(l))


def read_list(fn):
    with open(fn, 'r') as f:
        l = f.read().splitlines()
    return l


def compute_weights(distances, neighborhood_factor=3):
    from math import ceil
    radius_ii = ceil(distances.shape[1] / neighborhood_factor)
    sigma = distances[:, [radius_ii - 1]]
    sigma[sigma == 0] = 1
    weights = np.exp(-1 * distances ** 2 / sigma ** 2)
    wnorm = weights.sum(axis=1, keepdims=True)
    wnorm[wnorm == 0] = 1.0
    weights = weights / wnorm
    return weights


def neighbors_and_weights(data,
                          latent_obsm_key="spatial",
                          n_neighbors=30,
                          neighborhood_factor=3):
    """
    :param data:
    :param latent_obsm_key:
    :param n_neighbors:
    :param neighborhood_factor:
    :param approx_neighbors:
    :return:
    """
    from sklearn.neighbors import NearestNeighbors
    coords = data.obsm[latent_obsm_key]
    # get Nearest n Neighbors
    nbrs = NearestNeighbors(n_neighbors=n_neighbors, algorithm="ball_tree").fit(coords)
    dist, ind = nbrs.kneighbors()  # neighbor name and index
    # get spatial weights between two points
    weights = compute_weights(
        dist, neighborhood_factor=neighborhood_factor)
    ind_df = pd.DataFrame(ind, index=data.obs_names)
    neighbors = ind_df
    weights = pd.DataFrame(weights, index=neighbors.index,
                           columns=neighbors.columns)
    return ind, neighbors, weights


def get_neighbor_weight_matrix(df):
    """
    3 columns weight matrix, first column is cell, second column is neighbor cells, third column is weight values.
    Create {cell: [neighbors]} index dictionary
    and {cell: [weights]} value dictionary
    :param df:
    :return: neighbor index dict and neighbor weight dict for pysal.lib to create a weights.W object.
    """
    unique_cells = sorted(df['Cell_x'].unique())
    cell_to_index = {cell: idx for idx, cell in enumerate(unique_cells)}

    df['Cell_x'] = df['Cell_x'].map(cell_to_index)
    df['Cell_y'] = df['Cell_y'].map(cell_to_index)

    nei_dict = (df.groupby('Cell_x')['Cell_y'].apply(list).to_dict())

    weights_grouped = df.groupby('Cell_x')['Weight'].apply(list).to_dict()
    w_dict = {cell: weights_grouped[cell] for cell in nei_dict}
    return nei_dict, w_dict


def get_w(ind, weights_n):
    """Create a Weight object for esda program"""
    nind = pd.DataFrame(data=ind)
    nei = nind.transpose().to_dict('list')
    w_dict = weights_n.reset_index(drop=True).transpose().to_dict('list')
    from pysal.lib import weights
    w = weights.W(nei, weights=w_dict)
    return w


def flat_weights(cell_names, ind, weights, n_neighbors=30):
    """
    Turn neighbor index into
    :param cell_names:
    :param ind:
    :param weights:
    :param n_neighbors:
    :return:
    """
    cell1 = np.repeat(cell_names, n_neighbors)
    cell2_indices = ind.flatten()  # starts from 0
    cell2 = cell_names[cell2_indices]
    print(len(cell1))
    print(len(cell2))
    weight = weights.to_numpy().flatten()
    df = pd.DataFrame({
        "Cell_x": cell1,
        "Cell_y": cell2,
        "Weight": weight
    })
    return df


def square_weights(flat_weights_matrix):
    full_weights_matrix = pd.pivot_table(flat_weights_matrix,
                                         index='Cell_x',
                                         columns='Cell_y',
                                         values='Weight',
                                         fill_value=0)
    return full_weights_matrix


def fdr(ps, axis=0):
    """
    Apply the Benjamini-Hochberg procedure (FDR) of an array of p-values
    :param ps:
    :param axis:
    :return:
    """
    ps = np.asarray(ps)
    ps_in_range = (np.issubdtype(ps.dtype, np.number)
                   and np.all(ps == np.clip(ps, 0, 1)))
    if not ps_in_range:
        raise ValueError("`ps` must include only numbers between 0 and 1.")
    if axis is None:
        axis = 0
        ps = ps.ravel()
    axis = np.asarray(axis)[()]
    if not np.issubdtype(axis.dtype, np.integer) or axis.size != 1:
        raise ValueError("`axis` must be an integer or `None`")
    if ps.size <= 1 or ps.shape[axis] <= 1:
        return ps[()]
    ps = np.moveaxis(ps, axis, -1)
    m = ps.shape[-1]
    order = np.argsort(ps, axis=-1)
    ps = np.take_along_axis(ps, order, axis=-1)  # this copies ps
    # Equation 1 of [1] rearranged to reject when p is less than specified q
    i = np.arange(1, m + 1)
    ps *= m / i
    np.minimum.accumulate(ps[..., ::-1], out=ps[..., ::-1], axis=-1)
    # Restore original order of axes and data
    np.put_along_axis(ps, order, values=ps.copy(), axis=-1)
    ps = np.moveaxis(ps, -1, axis)
    return np.clip(ps, 0, 1)


def cal_s0(w):
    """
    s0=\sum_i \sum_j w_{i,j}
    """
    s0 = np.sum(w)
    return s0


def cal_s1(w):
    """
    s1 = 1/2 * sum_i sum_j (w_ij + w_ji)^2
    """
    # broadcast, create w_ij + w_ji 矩阵
    w_sum = w + w.T
    # 计算 s1
    s1 = 0.5 * np.sum(w_sum ** 2)
    return s1


def cal_s2(w):
    """
    s2 = \sum_j (\sum_i w_{i,j} + \sum_i w_{j,i})^2
    """
    # 计算行和列的和
    row_sums = np.sum(w, axis=1)
    col_sums = np.sum(w, axis=0)
    # 计算 \sum_i w_{i,j} + \sum_i w_{j,i} 对每个 j
    total_sums = row_sums + col_sums
    s2 = np.sum(total_sums ** 2)
    return s2


def format_gene_array(gene_array):
    if scipy.sparse.issparse(gene_array):
        gene_array = gene_array.toarray()
    return gene_array.reshape(-1)


def cal_k(gene_expression_matrix:np.ndarray, gene_x_id, n):
    gene_x_exp_mean = gene_expression_matrix[:, gene_x_id].mean()
    gene_x_exp = format_gene_array(gene_expression_matrix[:, gene_x_id])
    denominator = np.square(np.sum(np.square(gene_x_exp - gene_x_exp_mean)))
    numerator = n * np.sum(np.power(gene_x_exp - gene_x_exp_mean, 4))
    K = numerator / denominator
    return K


# -----------------------------------------------------#
# SOMDE
# -----------------------------------------------------#
def somde_p_values(adata, k=20, layer_key='raw_counts', latent_obsm_key="spatial"):
    if layer_key:
        exp = adata.layers[layer_key].toarray() if scipy.sparse.issparse(adata.layers[layer_key]) else adata.layers[layer_key]
    else:
        exp = adata.X.toarray() if scipy.sparse.issparse(adata.X) else adata.X
    df = pd.DataFrame(data=exp.T,
                      columns=adata.obs_names,
                      index=adata.var_names)
    cell_coordinates = adata.obsm[latent_obsm_key]
    corinfo = pd.DataFrame({
        "x": cell_coordinates[:, 0],
        "y": cell_coordinates[:, 1],
    })
    corinfo.index = adata.obs_names
    corinfo["total_count"] = exp.sum(1)
    X = corinfo[['x', 'y']].values.astype(np.float32)
    from somde import SomNode
    som = SomNode(X, k)
    som.mtx(df)
    som.norm()
    result, SVnum = som.run()
    p_values = result.pval
    adjusted_p_values = fdr(p_values)
    som.view()
    return adjusted_p_values


def view(som, raw=True, c=False, line=False):
    import matplotlib.pyplot as plt
    rr = som.som.codebook
    sizenum = np.ones([rr.shape[0], rr.shape[1]]) * 30
    rr = np.reshape(rr, [-1, 2])
    if raw:
        plt.scatter(som.X[:, 0], som.X[:, 1], s=3, label='original')
    for i in range(som.X.shape[0]):
        v, u = som.som.bmus[i]
        sizenum[u, v] += 2
        if line:
            plt.plot([som.X[i, 0], som.som.codebook[u, v, 0]], [som.X[i, 1], som.som.codebook[u, v, 1]])
    sizenum = np.reshape(sizenum, [-1, ])
    if c:
        plt.scatter(rr[:, 0], rr[:, 1], s=sizenum, label=str(som.somn) + 'X' + str(som.somn) + ' SOM nodes',
                    c=sizenum, cmap='hot')
        plt.colorbar()
    else:
        plt.scatter(rr[:, 0], rr[:, 1], s=sizenum, label=str(som.somn) + 'X' + str(som.somn) + ' SOM nodes', c='r')
    plt.savefig('somde.png')
    plt.close()


# -----------------------------------------------------#
# Main
# -----------------------------------------------------#
# def spatial_autocorrelation(adata,
#                             layer_key="raw_counts",
#                             latent_obsm_key="spatial",
#                             n_neighbors=10,
#                             somde_k=20,
#                             n_processes=None,
#                             prefix='',
#                             output_dir='.'):
#     """
#     Calculate spatial autocorrelation values using Moran's I, Geary'C, Getis's G and SOMDE algorithms
#     :param adata:
#     :param layer_key:
#     :param latent_obsm_key:
#     :param n_neighbors:
#     :param somde_k:
#     :param n_processes:
#     :param prefix:
#     :param output_dir:
#     :return:
#     """
#     print('Computing spatial weights matrix...')
#     sc.pp.filter_cells(adata, min_genes=100)
#     ind, neighbors, weights_n = neighbors_and_weights(adata, latent_obsm_key=latent_obsm_key, n_neighbors=n_neighbors)
#     Weights = get_w(ind, weights_n)
#     print("Computing Moran's I...")
#     # morans_ps = morans_i_p_values(adata, ind, weights_n, layer_key=layer_key, n_process=n_processes)
#     # # np.savetxt(f'{output_dir}/{prefix}_morans_ps.txt', morans_ps)
#     # fdr_morans_ps = fdr(morans_ps)
#     # # np.savetxt(f'{output_dir}/{prefix}_fdr_morans_ps.txt', fdr_morans_ps)
#     # print("Computing Geary's C...")
#     # gearys_cs = gearys_c_p_values(adata, ind, weights_n, layer_key=layer_key, n_process=n_processes)
#     # # np.savetxt(f'{output_dir}/{prefix}_gearys_cs.txt', gearys_cs)
#     # fdr_gearys_cs = fdr(gearys_cs)
#     # # np.savetxt(f'{output_dir}/{prefix}_fdr_gearys_cs.txt', fdr_gearys_cs)
#     # print("Computing Getis G...")
#     getis_gs = getis_g_p_values(adata, Weights, n_processes=n_processes, layer_key=layer_key)
#     np.savetxt(f'{output_dir}/{prefix}_getis_gs.txt', getis_gs)
#     # fdr_getis_gs = fdr(getis_gs)
#     # np.savetxt(f'{output_dir}/{prefix}_fdr_getis_gs.txt', fdr_getis_gs)
#     morans_ps = np.loadtxt(f'{output_dir}/{prefix}_morans_ps.txt')
#     fdr_morans_ps = np.loadtxt(f'{output_dir}/{prefix}_fdr_morans_ps.txt')
#     gearys_cs = np.loadtxt(f'{output_dir}/{prefix}_gearys_cs.txt')
#     fdr_gearys_cs = np.loadtxt(f'{output_dir}/{prefix}_fdr_gearys_cs.txt')
#     # getis_gs = np.loadtxt(f'{output_dir}/{prefix}_getis_gs.txt')
#     # fdr_getis_gs = np.loadtxt(f'{output_dir}/{prefix}_fdr_getis_gs.txt')
#     print('Computing SOMDE...')
#     adjusted_p_values = somde_p_values(adata, k=somde_k, layer_key=layer_key, latent_obsm_key=latent_obsm_key)
#     # np.savetxt(f'{output_dir}/{prefix}_fdr_SOMDE.txt', adjusted_p_values)
#     more_stats = pd.DataFrame({
#         'C': gearys_cs,
#         'FDR_C': fdr_gearys_cs,
#         'I': morans_ps,
#         'FDR_I': fdr_morans_ps,
#         'G': getis_gs,
#         'FDR_G': fdr_getis_gs,
#         'FDR_SOMDE': adjusted_p_values
#     }, index=adata.var_names)
#     more_stats.to_csv(f'{output_dir}/{prefix}_more_stats.csv', sep='\t')
#     return more_stats
#
#
# def combind_fdrs(pvalue_df, method='fisher') -> np.array:
#     """method options are {}"""
#     from scipy.stats import combine_pvalues
#     combined = np.apply_along_axis(combine_pvalues, 1, pvalue_df, method=method)[:, 1]
#     return combined  # shape (n_gene, )
#
#
# def preprocess(adata: ad.AnnData, min_genes=0, min_cells=3, min_counts=1, max_gene_num=4000):
#     adata.var_names_make_unique()  # compute the number of genes per cell (computes ‘n_genes' column)
#     # # find mito genes
#     sc.pp.ﬁlter_cells(adata, min_genes=0)
#     # add the total counts per cell as observations-annotation to adata
#     adata.obs['n_counts'] = np.ravel(adata.X.sum(axis=1))
#     # ﬁltering with basic thresholds for genes and cells
#     sc.pp.ﬁlter_cells(adata, min_genes=min_genes)
#     sc.pp.ﬁlter_genes(adata, min_cells=min_cells)
#     sc.pp.ﬁlter_genes(adata, min_counts=min_counts)
#     adata = adata[adata.obs['n_genes'] < max_gene_num, :]
#     return adata
#
#
# def hot(data, layer_key="raw_counts", latent_obsm_key="spatial"):
#     import hotspot
#     hs = hotspot.Hotspot(data,
#                          layer_key=layer_key,
#                          model='bernoulli',
#                          latent_obsm_key=latent_obsm_key)
#     hs.create_knn_graph(weighted_graph=False, n_neighbors=10)
#     hs_results = hs.compute_autocorrelations()
#     return hs, hs_results
#
#
# def select_genes(more_stats, hs_results, fdr_threshold=0.05, combine=True):
#     """
#     Select genes...
#     :param more_stats:
#     :param hs_results:
#     :param fdr_threshold:
#     :param combine: To select genes, combine p-values then choose genes have
#     :return:
#     """
#     # if combine:
#     more_stats['FDR'] = hs_results.FDR
#     cfdrs = combind_fdrs(more_stats[['FDR_C', 'FDR_I', 'FDR_G', 'FDR_SOMDE', 'FDR']])
#     more_stats['combined'] = cfdrs
#     cgenes = more_stats.loc[more_stats['combined'] < fdr_threshold].index
#     print(f"Combinded FDRs gives: {len(cgenes)} genes")  # 6961
#     # return genes
#     # else:
#     moran_genes = more_stats.loc[more_stats.FDR_I < fdr_threshold].index
#     geary_genes = more_stats.loc[more_stats.FDR_C < fdr_threshold].index
#     getis_genes = more_stats.loc[more_stats.FDR_G < fdr_threshold].index
#     somde_genes = more_stats.loc[more_stats.FDR_SOMDE < fdr_threshold].index
#     hs_genes = hs_results.loc[(hs_results.FDR < fdr_threshold)].index
#     inter_genes = set.intersection(set(moran_genes), set(geary_genes), set(getis_genes), set(somde_genes))
#     print(f"Moran's I find {len(moran_genes)} genes")  # 3860
#     print(f"Geary's C find {len(geary_genes)} genes")  # 7063
#     print(f'getis find {len(getis_genes)} genes')  # 4285
#     print(f'SOMDE find {len(somde_genes)} genes')  # 2593
#     print(f'intersection gene num: {len(inter_genes)}')  # 3416
#     # check somde results
#     t = set(somde_genes).intersection(set(hs_genes))
#     print(f'SOMDE genes {len(t)} in hs_genes')
#     # select
#     genes = inter_genes.intersection(set(hs_genes))
#     print(f'4 methods: inter_genes intersection with FDR genes: {len(genes)}')
#     global_genes = set.intersection(set(moran_genes), set(geary_genes), set(getis_genes)).intersection(set(hs_genes))
#     print(f'Global inter_genes intersection with FDR genes: {len(genes)}')  # 2728
#     return genes
#
#
# def main(prefix,
#          fn,
#          output_dir,
#          n_process=4,
#          min_genes=0,
#          min_cells=3,
#          min_counts=1,
#          layer_key=None,
#          latent_obsm_key="spatial",
#          n_neighbors=10,
#          somde_k=20):
#     """
#     Main function to calculate spatial autocorrelation values for genes
#     :param prefix: project name
#     :param fn: h5ad file name
#     :param output_dir: output directory
#     :param n_process: number of processes when computing in parallel
#     :param min_genes: filter cells
#     :param min_cells: filter genes
#     :param min_counts: filter genes
#     :param layer_key: layers containing gene expression matrix
#     :param latent_obsm_key: key stored cell/spot spatial coordinates
#     :param n_neighbors: number of neighbors when building KNN
#     :param somde_k: number of neighbors when using SOMDE model
#     :return:
#     """
#     adata = sc.read_h5ad(fn)
#     adata = preprocess(adata, min_genes=min_genes, min_cells=min_cells, min_counts=min_counts)
#     # ---- HOTSPOT autocorrelation
#     hs, hs_results = hot(adata, layer_key=layer_key, latent_obsm_key=latent_obsm_key)
#     # Select genes
#     more_stats = spatial_autocorrelation(adata,
#                                          layer_key=layer_key,
#                                          latent_obsm_key=latent_obsm_key,
#                                          n_neighbors=n_neighbors,
#                                          somde_k=somde_k,
#                                          n_processes=n_process,
#                                          prefix=prefix,
#                                          output_dir=output_dir)
#     select_genes(more_stats, hs_results, fdr_threshold=0.05, combine=False)
#
#
# if __name__ == '__main__':
#     project_id = sys.argv[1]
#     # 1. E14-16h
#     # more_stats = spatial_autocorrelation(sub_adata,
#     #                                      layer_key="raw_counts",
#     #                                      latent_obsm_key="spatial",
#     #                                      n_neighbors=10,
#     #                                      somde_k=20,
#     #                                      n_processes=None,
#     #                                      prefix='',
#     #                                      output_dir='.')
#     # select_genes(more_stats, hs_results, fdr_threshold=0.05, combine=False)
#     if project_id == 'E14':
#         fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/DATA/fly_pca/E14-16h_pca.h5ad'
#         output_dir = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/E14-16h'
#         prefix = 'E14-16h'
#         print(f'Running for {prefix} project...')
#         main(prefix, fn, output_dir, n_process=3, layer_key="raw_counts",latent_obsm_key="spatial",n_neighbors=10,somde_k=20)
#     elif project_id == 'E16':
#         fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/DATA/fly_pca/E16-18h_pca.h5ad'
#         output_dir = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/E16-18h'
#         prefix = 'E16-18h'
#         print(f'Running for {prefix} project...')
#         main(prefix, fn, output_dir, n_process=3, layer_key="raw_counts", latent_obsm_key="spatial", n_neighbors=10, somde_k=20)
#     elif project_id == 'L1':
#         fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/DATA/fly_pca/L1_pca.h5ad'
#         output_dir = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/L1'
#         prefix = 'L1'
#         print(f'Running for {prefix} project...')
#         main(prefix, fn, output_dir, n_process=3, layer_key="raw_counts", latent_obsm_key="spatial", n_neighbors=10, somde_k=20)
#     elif project_id == 'L2':
#         fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/DATA/fly_pca/L2_pca.h5ad'
#         output_dir = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/L2'
#         prefix = 'L2'
#         print(f'Running for {prefix} project...')
#         main(prefix, fn, output_dir, n_process=3, layer_key="raw_counts", latent_obsm_key="spatial", n_neighbors=10, somde_k=20)
#     elif project_id == 'L3':
#         fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/DATA/fly_pca/L3_pca.h5ad'
#         output_dir = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/L3'
#         prefix = 'L3'
#         print(f'Running for {prefix} project...')
#         main(prefix, fn, output_dir, n_process=3, layer_key="raw_counts", latent_obsm_key="spatial", n_neighbors=10, somde_k=20)
#
#     project_id = int(project_id)
#     # 1. dryad.8t8s248, MERFISH, 2024-07-16
#     # mouse brain, preoptic region
#     if project_id == 1:
#         fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/1.merfish/Moffitt_and_Bambah-Mukku_et_al_merfish_all_cells.h5ad'
#         output_dir = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/1.merfish'
#         prefix = 'merfish'
#         print(f'Running for {prefix} project...')
#         main(prefix, fn, output_dir, n_process=20, layer_key=None,latent_obsm_key="spatial",n_neighbors=10,somde_k=20)
#
#     # 2. zenodo.7551712, 10X Visium, human colorectal cancer (CRC)
#     # 7 individuals; two samples per patient
#     # GRN: using known TF-target interactions from DoRothEA
#     # GRCh38
#     elif project_id == 2:
#         fn = '/zfsqd1/ST_OCEAN/USRS/hankai/database/SpaGRN/zenodo.7551712/DeconvolutionResults_ST_CRC_BelgianCohort/sp.h5ad'
#         output_dir = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/2.zenodo.7551712_BelgianCohort'
#         prefix = 'zenodo.7551712_BelgianCohort'
#         print(f'Running for [{prefix}] project...')
#         main(prefix, fn, output_dir, n_process=3, min_genes=10, min_cells=50, min_counts=10, layer_key=None, latent_obsm_key="spatial", n_neighbors=10, somde_k=20)
#     elif project_id == 3:
#         fn = '/zfsqd1/ST_OCEAN/USRS/hankai/database/SpaGRN/zenodo.7551712/DeconvolutionResults_ST_CRC_KoreanCohort/sp.h5ad'
#         output_dir = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/3.zenodo.7551712_KoreanCohort'
#         prefix = 'zenodo.7551712_KoreanCohort'
#         print(f'Running for [{prefix}] project...')
#         main(prefix, fn, output_dir, n_process=3, min_genes=10, min_cells=50, min_counts=10, layer_key=None,
#              latent_obsm_key="spatial", n_neighbors=10, somde_k=20)
#     elif project_id == 4:
#         fn = '/zfsqd1/ST_OCEAN/USRS/hankai/database/SpaGRN/zenodo.7551712/DeconvolutionResults_ST_CRC_LiverMetastasis/sp.h5ad'
#         output_dir = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/4.zenodo.7551712_LiverMetastasis'
#         prefix = 'zenodo.7551712_LiverMetastasis'
#         print(f'Running for [{prefix}] project...')
#         main(prefix, fn, output_dir, n_process=3, min_genes=10, min_cells=50, min_counts=10, layer_key=None,
#              latent_obsm_key="spatial", n_neighbors=10, somde_k=20)
#
#     # 3. human ovarian tumour
#     # 10X Visium (GSE211956) & CosMx (zenodo.8287970)
#     # High-grade serous ovarian tumours from 10 patients diagnosed with stage III-IV cancers
#     elif project_id == 5:
#         fn = '/zfsqd1/ST_OCEAN/USRS/hankai/database/SpaGRN/zenodo.7551712/DeconvolutionResults_ST_CRC_LiverMetastasis/sp.h5ad'
#         output_dir = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/5.zenodo.7551712_LiverMetastasis'
#         prefix = 'zenodo.7551712_LiverMetastasis'
#         print(f'Running for [{prefix}] project...')
#         main(prefix, fn, output_dir, n_process=3, min_genes=10, min_cells=50, min_counts=10, layer_key=None,
#              latent_obsm_key="spatial", n_neighbors=10, somde_k=20)
#     elif project_id == 6:
#         fn = '/zfsqd1/ST_OCEAN/USRS/hankai/database/SpaGRN/zenodo.7551712/DeconvolutionResults_ST_CRC_LiverMetastasis/sp.h5ad'
#         output_dir = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/5.zenodo.7551712_LiverMetastasis'
#         prefix = 'zenodo.7551712_LiverMetastasis'
#         print(f'Running for [{prefix}] project...')
#         main(prefix, fn, output_dir, n_process=3, min_genes=10, min_cells=50, min_counts=10, layer_key=None,
#              latent_obsm_key="spatial", n_neighbors=10, somde_k=20)
#
#     # 4. mouse, CosMx, brain with PFF induced PD
#     # 1 female and 2 male mic
#     # zenodo.10729766
#     elif project_id == 6:
#         fn = '/zfsqd1/ST_OCEAN/USRS/hankai/database/SpaGRN/zenodo.10729766/seurat_object_3mon.rds'
#         output_dir = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/6.zenodo.10729766_mouse_brain'
#         prefix = 'zenodo.10729766_mouse_brain'
#         print(f'Running for [{prefix}] project...')
#         main(prefix, fn, output_dir, n_process=3, min_genes=1, min_cells=3, min_counts=1, layer_key=None,
#              latent_obsm_key="spatial", n_neighbors=10, somde_k=20)
#
#     # 5. zenodo.8063124
#     # human, Clear cell renal cell carcinoma (ccRCC)
#     # 12 tumor sections and 2 NAT controls
#     elif project_id == 7:
#         fn = '/zfsqd1/ST_OCEAN/USRS/hankai/database/SpaGRN/zenodo.10729766/seurat_object_3mon.rds'
#         output_dir = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/7.zenodo.8063124_human'
#         prefix = 'zenodo.10729766_mouse_brain'
#         print(f'Running for [{prefix}] project...')
#         main(prefix, fn, output_dir, n_process=3, min_genes=1, min_cells=3, min_counts=1, layer_key=None,
#              latent_obsm_key="spatial", n_neighbors=10, somde_k=20)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date: Created on 31 Dec 2024 15:00
# @Author: Yao LI
# @File: SpaGRN/g_autocor.py
import os
import sys
import time

import scipy
import scanpy as sc
import numpy as np
import pandas as pd
import anndata as ad

from pynndescent import NNDescent
from scipy.stats import chi2, norm
from scipy import stats
from scipy.spatial.distance import pdist, squareform
from scipy.sparse import csr_matrix, issparse

from esda.getisord import G
from esda.moran import Moran
from esda.geary import Geary

import multiprocessing
from tqdm import tqdm


# -----------------------------------------------------#
# Getis Ord General G
# -----------------------------------------------------#
# def _getis_g(x, w):
#     x = np.asarray(x)
#     w = np.asarray(w)
#     numerator = np.sum(np.sum(w * np.outer(x, x)))
#     denominator = np.sum(np.outer(x, x))
#     G = numerator / denominator
#     return G
#
#
# def _getis_g_p_value_one_gene(G, w, x):
#     n = w.shape[0]
#     s0 = cal_s0(w)
#     s02 = s0 * s0
#     s1 = cal_s1(w)
#     b0 = (n2 - 3 * n + 3) * s1 - n * s2 + 3 * s02
#     b1 = (-1.0) * ((n2 - n) * s1 - 2 * n * s2 + 6 * s02)
#     b2 = (-1.0) * (2 * n * s1 - (n + 3) * s2 + 6 * s02)
#     b3 = 4 * (n - 1) * s1 - 2 * (n + 1) * s2 + 8 * s02
#     b4 = s1 - s2 + s02
#     EG = s0 / (n * (n - 1))
#     numerator = b0 * (np.square(np.sum(x ** 2))) + b1 * np.sum(np.power(x, 4)) + b2 * np.square(np.sum(x)) * np.sum(
#         x ** 2) + b3 * np.sum(x) * np.sum(np.power(x, 3)) + b4 * np.power(np.sum(x), 4)
#     denominator = np.square((np.square(np.sum(x)) - np.sum(x ** 2))) * n * (n - 1) * (n - 2) * (n - 3)
#     VG = numerator / denominator - np.square(EG)
#     Z = (G - EG) / np.sqrt(VG)
#     p_value = 1 - norm.cdf(Z)
#     # print(f'G: {G}\nVG: {VG}\nZ: {Z}\np_value: {p_value}')
#     return p_value
#
#
# def getis_g_p_values_one_gene(gene_expression_matrix, gene_x_id, w):
#     g = G(gene_expression_matrix[:, gene_x_id], w)
#     return g.p_norm


# parallel computing
# p-values
def _compute_g_for_gene(args):
    gene_expression_matrix, gene_x_id, w = args
    g = G(gene_expression_matrix[:, gene_x_id], w)
    # print(f'gene{gene_x_id}: p_value: {g.p_norm}')
    return g.p_norm


def _getis_g_parallel(gene_expression_matrix, w, n_genes, n_processes=None):
    pool_args = [(gene_expression_matrix, gene_x_id, w) for gene_x_id in range(n_genes)]
    with multiprocessing.Pool(processes=n_processes) as pool:
        Gp_values = pool.map(_compute_g_for_gene, pool_args)
    return np.array(Gp_values)


# z-scores
def _compute_g_zscore_for_gene(args):
    gene_expression_matrix, gene_x_id, w = args
    g = G(gene_expression_matrix[:, gene_x_id], w)
    # print(f'gene{gene_x_id}: z_score: {g.z_norm}')
    return g.z_norm


def _getis_g_zscore_parallel(gene_expression_matrix, w, n_genes, n_processes=None):
    pool_args = [(gene_expression_matrix, gene_x_id, w) for gene_x_id in range(n_genes)]
    with multiprocessing.Pool(processes=n_processes) as pool:
        Gp_values = pool.map(_compute_g_zscore_for_gene, pool_args)
    return np.array(Gp_values)


def getis_g(adata,
            Weights,
            n_processes=None,
            layer_key=None,
            mode='pvalue'):
    """
    Calculate getis ord general g for all genes and return getis_g_p_values values as a numpy.array
    :param adata: data containing gene expression matrix and cell-feature spatial coordinates array
    :param Weights:
    :param n_processes: number of jobs when computing parallelly
    :param layer_key:
    :param mode: to calculate p-values or z-scores for the statistics
    :return: (numpy.array) dimension: (n_genes, )
    """
    assert mode in ['pvalue', 'zscore']
    n_genes = len(adata.var_names)
    # prepare gene expression matrix format
    if layer_key:
        gene_expression_matrix = adata.layers[layer_key].toarray() if scipy.sparse.issparse(
            adata.layers[layer_key]) else adata.layers[layer_key]
    else:
        gene_expression_matrix = adata.X.toarray() if scipy.sparse.issparse(adata.X) else adata.X

    # start calculation
    if mode == 'pvalue':
        p_values = _getis_g_parallel(gene_expression_matrix, Weights, n_genes, n_processes=n_processes)
        return p_values
    elif mode == 'zscore':
        z_scores = _getis_g_zscore_parallel(gene_expression_matrix, Weights, n_genes, n_processes=n_processes)
        return z_scores



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date: Created on 30 Jun 2024 23:01
# @Author: Yao LI
# @File: SpaGRN/corexp.py
import os
import sys
import time
import scanpy as sc
import numpy as np
import pandas as pd
import anndata as ad
import scipy
from scipy.sparse import csr_matrix, issparse
import multiprocessing
from tqdm import tqdm


# --------------------- CO-EXPRESSION --------------------------
# in case sparse X in h5ad
def format_gene_array(gene_array):
    if scipy.sparse.issparse(gene_array):
        gene_array = gene_array.toarray()
    return gene_array.reshape(-1)

class D:
    def __init__(self):
        pass
# Bivariate Moran's R
def bv_moran_r(gene_x_id, gene_y_id, cell_x_id, cell_y_id, wij, mtx):
    """Compute bivariate Moran's R value of two given genes"""
    # 3 get average
    gene_x_exp_mean = mtx[:, gene_x_id].mean()
    gene_y_exp_mean = mtx[:, gene_y_id].mean()
    # 4 calculate denominator
    gene_x_exp = format_gene_array(mtx[:, gene_x_id])
    gene_y_exp = format_gene_array(mtx[:, gene_y_id])
    denominator = np.sqrt(np.square(gene_x_exp - gene_x_exp_mean).sum()) * \
                  np.sqrt(np.square(gene_y_exp - gene_y_exp_mean).sum())
    # 5 calulate numerator
    gene_x_in_cell_i = format_gene_array(mtx[cell_x_id, gene_x_id])
    gene_y_in_cell_j = format_gene_array(mtx[cell_y_id, gene_y_id])
    numerator = np.sum(wij * (gene_x_in_cell_i - gene_x_exp_mean) * (gene_y_in_cell_j - gene_y_exp_mean))
    return numerator / denominator


def compute_pairs_m(args):
    """Apply function on two genes"""
    gene_names, gene_x_id, gene_y_id, cell_x_id, cell_y_id, wij, mtx = args
    value = bv_moran_r(gene_x_id, gene_y_id, cell_x_id, cell_y_id, wij, mtx)
    tf = gene_names[gene_x_id]
    gene = gene_names[gene_y_id]
    return tf, gene, value


def global_bivariate_moran_R(adata, weights: pd.DataFrame, tfs_in_data: list, select_genes: list, layer_key='raw_counts', num_workers: int = 4) -> pd.DataFrame:
    """
    :param adata:
    :param weights:
    :param tfs_in_data:
    :param select_genes:
    :param layer_key:
    :param num_workers:
    :return:
    """
    gene_names = adata.var.index
    # 1 gene name to matrix id
    tf_ids = adata.var.index.get_indexer(tfs_in_data)
    target_ids = adata.var.index.get_indexer(select_genes)
    # 2 cell name to matrix id
    tmp_obs = adata.obs.copy()
    tmp_obs['id'] = np.arange(len(adata.obs))
    cell_x_id = tmp_obs.loc[weights['Cell_x'].to_list()]['id'].to_numpy()
    cell_y_id = tmp_obs.loc[weights['Cell_y'].to_list()]['id'].to_numpy()
    if layer_key:
        mtx = adata.layers[layer_key].toarray() if scipy.sparse.issparse(adata.layers[layer_key]) else adata.layers[
            layer_key]
    else:
        mtx = adata.X.toarray() if scipy.sparse.issparse(adata.X) else adata.X
    wij = weights['Weight'].to_numpy()
    start_time = time.time()
    pool_args = [(gene_names, gene_x_id, gene_y_id, cell_x_id, cell_y_id, wij, mtx) for gene_x_id in tf_ids for gene_y_id in target_ids]
    with multiprocessing.Pool(processes=num_workers) as pool:
        results = pool.map(compute_pairs_m, pool_args)
    results_df = pd.DataFrame(results, columns=['TF', 'target', 'importance'])
    end_time = time.time()
    total_time = end_time - start_time
    print(f"run_in_parallel_m: Total time taken: {total_time:.4f} seconds")
    return results_df


# Bivariate Greay's C
def bv_geary_c(gene_x_id, gene_y_id, cell_x_id, cell_y_id, wij, mtx):
    """Compute bivariate Geary's C value of two given genes"""
    gene_x_exp_mean = mtx[:, gene_x_id].mean()
    gene_y_exp_mean = mtx[:, gene_y_id].mean()
    # 4 calculate denominator
    gene_x_exp = format_gene_array(mtx[:, gene_x_id])
    gene_y_exp = format_gene_array(mtx[:, gene_y_id])
    denominator = np.sqrt(np.square(gene_x_exp - gene_x_exp_mean).sum()) * \
                  np.sqrt(np.square(gene_y_exp - gene_y_exp_mean).sum())
    # 5 calulate numerator
    gene_x_in_cell_i = format_gene_array(mtx[cell_x_id, gene_x_id])
    gene_x_in_cell_j = format_gene_array(mtx[cell_y_id, gene_x_id])
    gene_y_in_cell_i = format_gene_array(mtx[cell_x_id, gene_y_id])
    gene_y_in_cell_j = format_gene_array(mtx[cell_y_id, gene_y_id])
    numerator = np.sum(wij * (gene_x_in_cell_i - gene_x_in_cell_j) * (gene_y_in_cell_i - gene_y_in_cell_j))
    return numerator / denominator


def compute_pairs_c(args):
    """Apply function on two genes"""
    gene_names, gene_x_id, gene_y_id, cell_x_id, cell_y_id, wij, mtx = args
    value = bv_geary_c(gene_x_id, gene_y_id, cell_x_id, cell_y_id, wij, mtx)
    tf = gene_names[gene_x_id]
    gene = gene_names[gene_y_id]
    return tf, gene, value


def global_bivariate_gearys_C(adata, weights: pd.DataFrame, tfs_in_data: list, select_genes: list, layer_key='raw_counts', num_workers: int = 4) -> pd.DataFrame:
    """
    :param adata:
    :param weights:
    :param tfs_in_data:
    :param select_genes:
    :param layer_key:
    :param num_workers:
    :return:
    """
    gene_names = adata.var.index
    # 1 gene name to matrix id
    tf_ids = adata.var.index.get_indexer(tfs_in_data)
    target_ids = adata.var.index.get_indexer(select_genes)
    # 2 cell name to matrix id
    tmp_obs = adata.obs.copy()
    tmp_obs['id'] = np.arange(len(adata.obs))
    cell_x_id = tmp_obs.loc[weights['Cell_x'].to_list()]['id'].to_numpy()
    cell_y_id = tmp_obs.loc[weights['Cell_y'].to_list()]['id'].to_numpy()
    # 3 get average
    if layer_key:
        mtx = adata.layers[layer_key].toarray() if scipy.sparse.issparse(adata.layers[layer_key]) else adata.layers[layer_key]
    else:
        mtx = adata.X.toarray() if scipy.sparse.issparse(adata.X) else adata.X
    wij = weights['Weight'].to_numpy()
    start_time = time.time()
    pool_args = [(gene_names, gene_x_id, gene_y_id, cell_x_id, cell_y_id, wij, mtx) for gene_x_id in tf_ids for gene_y_id in target_ids]
    with multiprocessing.Pool(processes=num_workers) as pool:
        results = pool.map(compute_pairs_c, pool_args)
    results_df = pd.DataFrame(results, columns=['TF', 'target', 'importance'])
    end_time = time.time()
    total_time = end_time - start_time
    print(f"run_in_parallel_m: Total time taken: {total_time:.4f} seconds")
    return results_df


def preprocess(adata: ad.AnnData, min_genes=0, min_cells=3, min_counts=1, max_gene_num=4000):
    """
    Perform cleaning and quality control on the imported data before constructing gene regulatory network
    :param min_genes:
    :param min_cells:
    :param min_counts:
    :param max_gene_num:
    :return: a anndata.AnnData
    """
    adata.var_names_make_unique()  # compute the number of genes per cell (computes ‘n_genes' column)
    # # find mito genes
    sc.pp.ﬁlter_cells(adata, min_genes=0)
    # add the total counts per cell as observations-annotation to adata
    adata.obs['n_counts'] = np.ravel(adata.X.sum(axis=1))
    # ﬁltering with basic thresholds for genes and cells
    sc.pp.ﬁlter_cells(adata, min_genes=min_genes)
    sc.pp.ﬁlter_genes(adata, min_cells=min_cells)
    sc.pp.ﬁlter_genes(adata, min_counts=min_counts)
    adata = adata[adata.obs['n_genes'] < max_gene_num, :]
    return adata


def read_list(fn):
    with open(fn, 'r') as f:
        l = f.read().splitlines()
    return l


def compute_weights(distances, neighborhood_factor=3):
    from math import ceil
    radius_ii = ceil(distances.shape[1] / neighborhood_factor)
    sigma = distances[:, [radius_ii - 1]]
    sigma[sigma == 0] = 1
    weights = np.exp(-1 * distances ** 2 / sigma ** 2)
    wnorm = weights.sum(axis=1, keepdims=True)
    wnorm[wnorm == 0] = 1.0
    weights = weights / wnorm
    return weights


def neighbors_and_weights(data,
                          latent_obsm_key="spatial",
                          n_neighbors=10,
                          neighborhood_factor=3):
    """
    :param data:
    :param latent_obsm_key:
    :param n_neighbors:
    :param neighborhood_factor:
    :param approx_neighbors:
    :return:
    """
    from sklearn.neighbors import NearestNeighbors
    coords = data.obsm[latent_obsm_key]
    # get Nearest n Neighbors
    nbrs = NearestNeighbors(n_neighbors=n_neighbors, algorithm="ball_tree").fit(coords)
    dist, ind = nbrs.kneighbors()  # neighbor name and index
    # get spatial weights between two points
    weights = compute_weights(
        dist, neighborhood_factor=neighborhood_factor)
    ind_df = pd.DataFrame(ind, index=data.obs_names)
    neighbors = ind_df
    weights = pd.DataFrame(weights, index=neighbors.index,
                           columns=neighbors.columns)
    return ind, neighbors, weights


def flat_weights(cell_names, ind, weights, n_neighbors=30):
    """
    Turn neighbor index into
    :param cell_names:
    :param ind:
    :param weights:
    :param n_neighbors:
    :return:
    """
    cell1 = np.repeat(cell_names, n_neighbors)
    cell2_indices = ind.flatten()  # starts from 0
    cell2 = cell_names[cell2_indices]
    weight = weights.to_numpy().flatten()
    df = pd.DataFrame({
        "Cell_x": cell1,
        "Cell_y": cell2,
        "Weight": weight
    })
    return df


def get_p_M(I, adata, weights, gene_x, gene_y, permutation_num=99):
    def __calc(w, zx, zy, den):
        wzy = slag(w, zy)
        num = (zx * wzy).sum()
        return num / den
    gene_matrix = adata.X
    x = gene_matrix[:, gene_x]
    x = np.asarray(x).flatten()
    y = gene_matrix[:, gene_y]
    y = np.asarray(y).flatten()
    zx = (x - x.mean()) / x.std(ddof=1)
    zy = (y - y.mean()) / y.std(ddof=1)
    den = x.shape[0] - 1.0
    sim = [__calc(weights, zx, np.random.permutation(zy), den) for i in range(permutation_num)]
    sim = np.array(sim)
    above = sim >= I
    larger = above.sum()
    if (permutation_num - larger) < larger:
        larger = permutation_num - larger
    p_sim = (larger + 1.0) / (permutation_num + 1.0)
    return p_sim


def get_p_C(C, adata, weights, gene_x, gene_y, permutation_num=99):
    sim = [bv_geary_c(adata, weights, gene_x, np.random.permutation(gene_y)) for i in range(permutation_num)]
    sim = np.array(sim)
    above = sim >= C
    larger = above.sum()
    if (permutation_num - larger) < larger:
        larger = permutation_num - larger
    p_sim = (larger + 1.0) / (permutation_num + 1.0)
    return p_sim


def main(data_fn, tfs_fn, genes_fn, layer_key='raw_counts', latent_obsm_key="spatial", n_neighbors=10, fw_fn=None, output_dir='.', num_workers=6):
    print('Loading experimental data...')
    adata = sc.read_h5ad(data_fn)
    adata = preprocess(adata)
    tfs = read_list(tfs_fn)
    select_genes = read_list(genes_fn)
    adata = adata[:, select_genes]
    tfs_in_data = list(set(tfs).intersection(set(adata.var_names)))
    print(f'{len(tfs_in_data)} TFs in data')
    select_genes_not_tfs = list(set(select_genes) - set(tfs_in_data))
    print(f'{len(select_genes_not_tfs)} genes to use.')
    # ---- Weights
    if fw_fn:
        fw = pd.read_csv(fw_fn)
    else:
        print('Computing spatial weights matrix...')
        ind, neighbors, weights_n = neighbors_and_weights(adata, latent_obsm_key=latent_obsm_key, n_neighbors=n_neighbors)
        cell_names = adata.obs_names
        print('Shifting spatial weight matrix shape...')
        fw = flat_weights(cell_names, ind, weights_n, n_neighbors=n_neighbors)
        fw.to_csv(f'{output_dir}/flat_weights.csv', index=False)
    # --- TEST BV
    print("Computing global bivariate geary'C value in parallel...")
    local_correlations_bv_gc = global_bivariate_gearys_C(adata,
                                                         fw,
                                                         tfs_in_data,
                                                         select_genes,
                                                         num_workers=num_workers,
                                                         layer_key=layer_key)
    local_correlations_bv_gc.to_csv(f'{output_dir}/local_correlations_bv_gc.csv', index=None)
    print("Computing global bivariate Moran's R value in parallel...")
    local_correlations_bv_mr = global_bivariate_moran_R(adata,
                                                        fw,
                                                        tfs_in_data,
                                                        select_genes,
                                                        num_workers=num_workers,
                                                        layer_key=layer_key)
    local_correlations_bv_mr.to_csv(f'{output_dir}/local_correlations_bv_mr.csv', index=None)


if __name__ == '__main__':
    project_id = sys.argv[1]
    if project_id == 'E14-16h':
        data_fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/DATA/fly_pca/E14-16h_pca.h5ad'
        tfs_fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/06.stereopy/resource/tfs/allTFs_dmel.txt'
        genes_fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/E14-16h/hotspot_og/hotspot_select_genes.txt'
        fw_fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/E14-16h/E14-16h_flat_weights.csv'
        output_dir = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/E14-16h'
        main(data_fn, tfs_fn, genes_fn, fw_fn=fw_fn, output_dir=output_dir, num_workers=6)
    elif project_id == 'mouse_brain':
        data_fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/DATA/Mouse_brain_cell_bin.h5ad'
        tfs_fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/06.stereopy/resource/tfs/allTFs_mm.txt'
        genes_fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/mouse_brain/global_genes.txt'
        fw_fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/mouse_brain/flat_weights.csv'
        output_dir = f'/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/13.revision/{project_id}'
        main(data_fn, tfs_fn, genes_fn, layer_key='counts', latent_obsm_key="spatial", n_neighbors=10, fw_fn=fw_fn,
             output_dir=output_dir, num_workers=20)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date: Created on 31 Oct 2023 15:19
# @Author: Yao LI
# @File: spagrn/results.py
import csv

# third party modules
import json
import pandas as pd
from pyscenic.export import export2loom

from .network import Network


def dict_to_df(json_fn):
    """

    :param json_fn:
    :return:
    """
    dic = json.load(open(json_fn))
    df = pd.DataFrame([(key, var) for (key, L) in dic.items() for var in L], columns=['TF', 'targets'])
    df.to_csv(f'{json_fn.strip(".json")}.csv', index=False)


class HandleNetwork(Network):
    # Handle data generate by SpaGRN
    def __init__(self, adata, modules_fn=None, regulons_fn=None):
        super().__init__()
        self.data = adata
        self.load_results(modules_fn=modules_fn, regulons_fn=regulons_fn)

    def regulons_to_csv(self, fn: str = 'regulon_list.csv'):
        """
        Save regulon_list (df2regulons output) into a csv file.
        :param fn:
        :return:
        """
        # self.regulon_dict = self.get_regulon_dict(self.regulon_list)
        # Optional: join list of target genes
        for key in self.regulon_dict.keys(): self.regulon_dict[key] = ";".join(self.regulon_dict[key])
        # Write to csv file
        with open(fn, 'w') as f:
            w = csv.writer(f)
            w.writerow(["Regulons", "Target_genes"])
            w.writerows(self.regulon_dict.items())

    def to_loom(self, fn: str = 'output.loom'):
        """
        Save GRN results in one loom file
        :param fn:
        :return:
        """
        export2loom(ex_mtx=self.matrix, auc_mtx=self.auc_mtx,
                    regulons=[r.rename(r.name.replace('(+)', ' (' + str(len(r)) + 'g)')) for r in self.regulons],
                    out_fname=fn)

    def to_cytoscape(self,
                     tf: str,
                     fn: str = 'cytoscape.txt'):
        """
        Save GRN result of one TF, into Cytoscape format for down stream analysis
        :param tf: one target TF name
        :param fn: output file name
        :return:

        Example:
            grn.to_cytoscape(regulons, adjacencies, 'Gnb4', 'Gnb4_cytoscape.txt')
        """
        # get TF data
        if self.regulons is None:
            raise ValueError("run load_results(regulon_fn) first")
        if self.regulon_dict is None:
            self.regulon_dict = self.get_regulon_dict(self.regulons)
        sub_adj = self.adjacencies[self.adjacencies.TF == tf]
        targets = self.regulon_dict[f'{tf}(+)']
        # all the target genes of the TF
        sub_df = sub_adj[sub_adj.target.isin(targets)]
        sub_df.to_csv(fn, index=False, sep='\t')

    def get_cytoscape(self,
                      tf: str):
        """
        Save GRN result of one TF, into Cytoscape format for down stream analysis
        :param tf: one target TF name
        :return:
        Example:
            grn.get_cytoscape(regulons, adjacencies, 'Gnb4')
        """
        tf = tf if '(+)' not in tf else tf.replace('(+)', '')
        # get TF data
        if self.regulons is None:
            raise ValueError("run load_results(regulon_fn) first")
        if self.regulon_dict is None:
            self.regulon_dict = self.get_regulon_dict(self.regulons)
        if f'{tf}(+)' not in self.regulon_dict:
            print(f'{tf}(+) not found in data.')
            return

        sub_adj = self.adjacencies[self.adjacencies.TF == tf]
        targets = self.regulon_dict[f'{tf}(+)']
        # all the target genes of the TF
        sub_df = sub_adj[sub_adj.target.isin(targets)]
        sub_df.to_csv(f'{tf}_cytoscape.txt', index=False, sep='\t')

    def get_metascape(self, tf):
        tf = tf if '(+)' not in tf else tf.replace('(+)', '')
        # get TF data
        if self.regulons is None:
            raise ValueError("run load_results(regulon_fn) first")
        if self.regulon_dict is None:
            self.regulon_dict = self.get_regulon_dict(self.regulons)
        if f'{tf}(+)' not in self.regulon_dict:
            print(f'{tf}(+) not found in data.')
            return
        targets = self.regulon_dict[f'{tf}(+)']
        with open(f'{tf}_metascape.txt', 'w') as f:
            f.writelines('\n'.join(targets))

    def shared_targets(self, reg1, reg2):
        """
        Find shared targets between regulon1 and regulon2, target genes are ranked by their importance value.
        :param reg1:
        :param reg2:
        :return:
        """
        pass

    def get_targets(self, reg):
        """
        Get target genes of a regulon and save them into a text file.
        :param reg:
        :return:
        """
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date: Created on 31 Dec 2024 15:01
# @Author: Yao LI
# @File: SpaGRN/m_autocor.py
import os
import sys
import time

import scipy
import scanpy as sc
import numpy as np
import pandas as pd
import anndata as ad

from pynndescent import NNDescent
from scipy.stats import chi2, norm
from scipy import stats
from scipy.spatial.distance import pdist, squareform
from scipy.sparse import csr_matrix, issparse

from esda.getisord import G
from esda.moran import Moran
from esda.geary import Geary

import multiprocessing
from tqdm import tqdm


# -----------------------------------------------------#
# M's I
# -----------------------------------------------------#
# def _morans_i(adata, weights, layer_key='raw_counts'):
#     if 'connectivities' not in adata.obsp.keys():
#         adata.obsp['connectivities'] = weights
#     morans_i_array = sc.metrics.morans_i(adata, layer=layer_key)
#     # shape: (n_genes, )
#     return morans_i_array
#
#
# def _morans_i_p_value_one_gene(adata, gene_x_id, weights, morans_i_array):
#     I = morans_i_array[gene_x_id]  # moran's I stats for the gene
#     n = len(adata.obs_names)  # number of cells
#     EI = -1 / (n - 1)  # Moran’s I expected value
#     K = cal_k(adata, gene_x_id, n)
#     S0 = cal_s0(weights)
#     S1 = cal_s1(weights)
#     S2 = cal_s2(weights)
#     # Variance
#     part1 = (n * (S1 * (n ** 2 - 3 * n + 3) - n * S2 + 3 * np.square(S0))) / (
#             (n - 1) * (n - 2) * (n - 3) * np.square(S0))
#     part2 = (K * (S1 * (n ** 2 - n) - 2 * n * S2 + 6 * np.square(S0))) / ((n - 1) * (n - 2) * (n - 3) * np.square(S0))
#     VI = part1 - part2 - np.square(EI)
#     stdI = np.sqrt(VI)
#     # Z score
#     Z = (I - EI) / stdI
#     return Z
#     # Perform one-tail test one z score
#     # p_value = 1 - norm.cdf(Z)  # right tail
#     # return p_value
#
#
# def morans_i_p_value_one_gene(x, w):
#     i = Moran(x, w)
#     return i.p_norm


# parallel computing
def _compute_i_for_gene(args):
    x, w = args
    i = Moran(x, w)
    return i.p_norm


def _compute_i_zscore_for_gene(args):
    x, w = args
    i = Moran(x, w)
    return i.z_norm


def _morans_i_parallel(n_genes, gene_expression_matrix, w, n_processes=None):
    pool_args = [(gene_expression_matrix[:, gene_x_id], w) for gene_x_id in range(n_genes)]
    with multiprocessing.Pool(processes=n_processes) as pool:
        p_values = pool.map(_compute_i_for_gene, pool_args)
    return np.array(p_values)


def _morans_i_zscore_parallel(n_genes, gene_expression_matrix, w, n_processes=None):
    pool_args = [(gene_expression_matrix[:, gene_x_id], w) for gene_x_id in range(n_genes)]
    with multiprocessing.Pool(processes=n_processes) as pool:
        z_scores = pool.map(_compute_i_zscore_for_gene, pool_args)
    return np.array(z_scores)


def morans_i_p_values(adata, Weights, layer_key='raw_counts', n_process=None):
    """
    Calculate Moran’s I Global Autocorrelation Statistic and its adjusted p-value
    :param adata: Anndata
    :param Weights:
    :param layer_key:
    :param n_process:
    :return:
    """
    n_genes = len(adata.var_names)
    # nind = pd.DataFrame(data=ind)
    # nei = nind.transpose().to_dict('list')
    # w_dict = weights_n.reset_index(drop=True).transpose().to_dict('list')
    # w = weights.W(nei, weights=w_dict)
    # w = get_w(ind, weights_n)
    if layer_key:
        gene_expression_matrix = adata.layers[layer_key].toarray() if scipy.sparse.issparse(adata.layers[layer_key]) else adata.layers[layer_key]
    else:
        gene_expression_matrix = adata.X.toarray() if scipy.sparse.issparse(adata.X) else adata.X
    p_values = _morans_i_parallel(n_genes, gene_expression_matrix, Weights, n_processes=n_process)
    return p_values


def morans_i_zscore(adata, Weights, layer_key='raw_counts', n_process=None):
    """
    Calculate Moran’s I Global Autocorrelation Statistic and its adjusted p-value
    :param adata: Anndata
    :param Weights:
    :param layer_key:
    :param n_process:
    :return:
    """
    n_genes = len(adata.var_names)
    if layer_key:
        gene_expression_matrix = adata.layers[layer_key].toarray() if scipy.sparse.issparse(adata.layers[layer_key]) else adata.layers[layer_key]
    else:
        gene_expression_matrix = adata.X.toarray() if scipy.sparse.issparse(adata.X) else adata.X
    z_scores = _morans_i_zscore_parallel(n_genes, gene_expression_matrix, Weights, n_processes=n_process)
    return z_scores



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date: Created on 15 Nov 2023 14:32
# @Author: Yao LI
# @File: spagrn/params.py


class InferParam:
    def __init__(self, method=None, rank_threshold=1500, auc_threshold=0.05, motif_similarity_fdr=0.05,
                 c_threshold=-1, layer_key=None, model='danb', latent_obsm_key='spatial', umi_counts_obs_key=None,
                 n_neighbors=30, weighted_graph=False, fdr_threshold=0.05, sigm=15,
                 nes_threshold=None, noweights=False, normalize=False, rho_mask_dropouts=False,
                 **kwargs):
        self.method = None
        self.set_method(method)
        self.rank_threshold = rank_threshold
        self.auc_threshold = auc_threshold
        self.motif_similarity_fdr = motif_similarity_fdr
        self.noweights = noweights
        self.c_threshold = c_threshold
        self.layer_key = layer_key
        self.model = model
        self.latent_obsm_key = latent_obsm_key
        self.umi_counts_obs_key = umi_counts_obs_key
        self.n_neighbors = n_neighbors
        self.weighted_graph = weighted_graph
        self.fdr_threshold = fdr_threshold
        self.sigm = sigm
        self.nes_threshold = nes_threshold
        self.normalize = normalize
        self.rho_mask_dropouts = rho_mask_dropouts

        # Store additional keyword arguments as instance variables
        for key, value in kwargs.items():
            setattr(self, key, value)

    def set_method(self, method):
        if method in ['spg', 'scc', 'grnboost']:
            self.method = method
        else:
            raise ValueError(f"Invalid method: {method}. Valid options are 'spg', 'scc', 'grnboost'.")

    def get_param(self, keyword, default=None):
        return getattr(self, keyword, default)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date: Created on 31 Dec 2024 15:02
# @Author: Yao LI
# @File: SpaGRN/c_autocor.py
import os
import sys
import time

import scipy
import scanpy as sc
import numpy as np
import pandas as pd
import anndata as ad

from pynndescent import NNDescent
from scipy.stats import chi2, norm
from scipy import stats
from scipy.spatial.distance import pdist, squareform
from scipy.sparse import csr_matrix, issparse

from esda.getisord import G
from esda.moran import Moran
from esda.geary import Geary

import multiprocessing
from tqdm import tqdm


# -----------------------------------------------------#
# G's C
# -----------------------------------------------------#
# def _gearys_c(adata, weights, layer_key='raw_counts'):
#     if 'connectivities' not in adata.obsp.keys():
#         adata.obsp['connectivities'] = weights
#     gearys_c_array = sc.metrics.gearys_c(adata, layer=layer_key)
#     # shape: (n_genes, )
#     return gearys_c_array
#
#
# def _gearys_c_p_value_one_gene(gene_expression_matrix, n_genes, gene_x_id, weights, gearys_c_array):
#     C = gearys_c_array[gene_x_id]
#     EC = 1
#     K = cal_k(gene_expression_matrix, gene_x_id, n_genes)
#     S0 = cal_s0(weights)
#     S1 = cal_s1(weights)
#     S2 = cal_s2(weights)
#     part1 = (n_genes - 1) * S1 * (n_genes ** 2 - 3 * n_genes + 3 - K * (n_genes - 1)) / (np.square(S0) * n_genes * (n_genes - 2) * (n_genes - 3))
#     part2 = (n_genes ** 2 - 3 - K * np.square(n_genes - 1)) / (n_genes * (n_genes - 2) * (n_genes - 3))
#     part3 = (n_genes - 1) * S2 * (n_genes ** 2 + 3 * n_genes - 6 - K * (n_genes ** 2 - n_genes + 2)) / (4 * n_genes * (n_genes - 2) * (n_genes - 3) * np.square(S0))
#     VC = part1 + part2 - part3
#     # variance = (2 * (n ** 2) * S1 - n * S2 + 3 * (S0 ** 2)) / (S0 ** 2 * (n - 1) * (n - 2) * (n - 3))
#     VC_norm = (1 / (2 * (n_genes + 1) * S0 ** 2)) * ((2 * S1 + S2) * (n_genes - 1) - 4 * S0 ** 2)
#     Z = (C - EC) / np.sqrt(VC_norm)
#     p_value = 1 - norm.cdf(Z)
#     print(f'C: {C}\nVC: {VC}\nVC_norm: {VC_norm}\nZ: {Z}\np_value: {p_value}')
#     return p_value


def gearys_c_p_value_one_gene(x, w):
    c = Geary(x, w)
    return c.p_norm


def gearys_c_z_score_one_gene(x, w):
    c = Geary(x, w)
    return c.z_norm


# parallel computing
def _compute_c_for_gene(args):
    x, w = args
    Cp = gearys_c_p_value_one_gene(x, w)
    return Cp


def _gearys_c_parallel(n_genes, gene_expression_matrix, w, n_processes=None):
    pool_args = [(gene_expression_matrix[:, gene_x_id], w) for gene_x_id in range(n_genes)]
    with multiprocessing.Pool(processes=n_processes) as pool:
        p_values = pool.map(_compute_c_for_gene, pool_args)
    return np.array(p_values)


def _compute_c_z_score_for_gene(args):
    x, w = args
    Cz = gearys_c_z_score_one_gene(x, w)
    return Cz


def _gearys_c_z_score_parallel(n_genes, gene_expression_matrix, w, n_processes=None):
    pool_args = [(gene_expression_matrix[:, gene_x_id], w) for gene_x_id in range(n_genes)]
    with multiprocessing.Pool(processes=n_processes) as pool:
        z_scores = pool.map(_compute_c_z_score_for_gene, pool_args)
    return np.array(z_scores)


def gearys_c(adata, Weights, layer_key='raw_counts', n_process=None, mode='pvalue'):
    """
    Main function to calculate Geary's C and its p-values
    :param adata:
    :param Weights:
    :param layer_key:
    :param n_process:
    :param mode
    :return:
    """
    n_genes = len(adata.var_names)
    if layer_key:
        gene_expression_matrix = adata.layers[layer_key].toarray() if scipy.sparse.issparse(adata.layers[layer_key]) else adata.layers[layer_key]
    else:
        gene_expression_matrix = adata.X.toarray() if scipy.sparse.issparse(adata.X) else adata.X
    if mode =='pvalue':
        p_values = _gearys_c_parallel(n_genes, gene_expression_matrix, Weights, n_processes=n_process)
        return p_values
    elif mode == 'zscore':
        z_scores = _gearys_c_z_score_parallel(n_genes, gene_expression_matrix, Weights, n_processes=n_process)
        return z_scores

